#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <objbase.h>
#include <commctrl.h>
#include <richedit.h>
#include "AkelEdit.h"
#include "AkelDLL.h"
#include "StackFunc.h"
#include "ConvFunc.h"
#include "ISystemFunction.h"
#include "IDocument.h"
#include "IServer.h"
#include "Scripts.h"
#include "Resources\Resource.h"


//Include conversion functions
#define xatoiW
#include "ConvFunc.h"

//Global variables
ITypeInfo *g_DocumentTypeInfo;
IDispatch *g_objCallback;
HWND g_hWndDialog=NULL;

const IDocumentVtbl MyIDocumentVtbl={
  Document_QueryInterface,
  Document_AddRef,
  Document_Release,
  Document_GetTypeInfoCount,
  Document_GetTypeInfo,
  Document_GetIDsOfNames,
  Document_Invoke,
  Document_GetMainWnd,
  Document_GetAkelDir,
  Document_GetInstanceDll,
  Document_IsOldWindows,
  Document_IsOldRichEdit,
  Document_IsOldComctl32,
  Document_IsAkelEdit,
  Document_IsMDI,
  Document_GetEditWnd,
  Document_GetEditFile,
  Document_GetEditCodePage,
  Document_GetEditBOM,
  Document_GetEditNewLine,
  Document_GetEditModified,
  Document_GetEditReadOnly,
  Document_SendMessage,
  Document_MessageBox,
  Document_InputBox,
  Document_GetSelStart,
  Document_GetSelEnd,
  Document_SetSel,
  Document_GetSelText,
  Document_ReplaceSel,
  Document_TextFind,
  Document_TextReplace,
  Document_GetClipboardText,
  Document_SetClipboardText,
  Document_Call,
  Document_IsPluginRunning,
  Document_OpenFile,
  Document_ReadFile,
  Document_SaveFile,
  Document_SystemFunction,
  Document_MemAlloc,
  Document_MemCopy,
  Document_MemRead,
  Document_MemFree,
  Document_WindowRegisterClass,
  Document_WindowUnregisterClass,
  Document_WindowGetMessage
};


//// IDocument

HRESULT STDMETHODCALLTYPE Document_QueryInterface(IDocument *this, REFIID vTableGuid, void **ppv)
{
  if (IsEqualIID(vTableGuid, &IID_IUnknown) || IsEqualIID(vTableGuid, &IID_IDocument) || IsEqualIID(vTableGuid, &IID_IDispatch))
  {
    *ppv=this;
    this->lpVtbl->Document_AddRef(this);
    return NOERROR;
  }
  *ppv=NULL;
  return E_NOINTERFACE;
}

ULONG STDMETHODCALLTYPE Document_AddRef(IDocument *this)
{
  return ++((IRealDocument *)this)->dwCount;
}

ULONG STDMETHODCALLTYPE Document_Release(IDocument *this)
{
  if (--((IRealDocument *)this)->dwCount == 0)
  {
    GlobalFree(this);
    InterlockedDecrement(&g_nObjs);
    return 0;
  }
  return ((IRealDocument *)this)->dwCount;
}


//// IDispatch

HRESULT Document_LoadRegTypeInfo(void)
{
  LPTYPELIB pTypeLib;
  HRESULT hr;

  if ((hr=LoadRegTypeLib(&CLSID_TypeLib, 1, 0, 0, &pTypeLib)) == S_OK)
  {
    if ((hr=pTypeLib->lpVtbl->GetTypeInfoOfGuid(pTypeLib, &IID_IDocument, &g_DocumentTypeInfo)) == S_OK)
    {
      pTypeLib->lpVtbl->Release(pTypeLib);
      g_DocumentTypeInfo->lpVtbl->AddRef(g_DocumentTypeInfo);
    }
  }
  return hr;
}

ULONG STDMETHODCALLTYPE Document_GetTypeInfoCount(IDocument *this, UINT *pCount)
{
  *pCount=1;
  return S_OK;
}

ULONG STDMETHODCALLTYPE Document_GetTypeInfo(IDocument *this, UINT itinfo, LCID lcid, ITypeInfo **pTypeInfo)
{
  HRESULT hr;

  *pTypeInfo=NULL;

  if (itinfo)
  {
    hr=ResultFromScode(DISP_E_BADINDEX);
  }
  else if (g_DocumentTypeInfo)
  {
    g_DocumentTypeInfo->lpVtbl->AddRef(g_DocumentTypeInfo);
    hr=S_OK;
  }
  else
  {
    hr=Document_LoadRegTypeInfo();
  }
  if (hr == S_OK) *pTypeInfo=g_DocumentTypeInfo;

  return hr;
}

ULONG STDMETHODCALLTYPE Document_GetIDsOfNames(IDocument *this, REFIID riid, LPOLESTR *rgszNames, UINT cNames, LCID lcid, DISPID *rgdispid)
{
  if (!g_DocumentTypeInfo)
  {
    HRESULT hr;

    if ((hr=Document_LoadRegTypeInfo()) != S_OK) return hr;
  }
  return DispGetIDsOfNames(g_DocumentTypeInfo, rgszNames, cNames, rgdispid);
}

ULONG STDMETHODCALLTYPE Document_Invoke(IDocument *this, DISPID dispid, REFIID riid, LCID lcid, WORD wFlags, DISPPARAMS *params, VARIANT *result, EXCEPINFO *pexcepinfo, UINT *puArgErr)
{
  if (!IsEqualIID(riid, &IID_NULL))
    return DISP_E_UNKNOWNINTERFACE;

  if (!g_DocumentTypeInfo)
  {
    HRESULT hr;

    if ((hr=Document_LoadRegTypeInfo()) != S_OK) return hr;
  }
  return DispInvoke(this, g_DocumentTypeInfo, dispid, wFlags, params, result, pexcepinfo, puArgErr);
}


//// IDocument methods

void IsAkelPadReady()
{
  if (bOldWindows)
  {
    if (!hMainWnd || !hAkelProcess)
    {
      MessageBoxA(NULL, GetLangStringA(wLangSystem, STRID_WINDOW_NOT_FOUND), "Scripts Plugin", MB_OK|MB_ICONEXCLAMATION);
      ExitProcess(0);
    }
    if (dwArchitectVersion == (DWORD)-1)
    {
      MessageBoxA(NULL, GetLangStringA(wLangSystem, STRID_ARCHITECTURE_MISMATCH), "Scripts plugin", MB_OK|MB_ICONEXCLAMATION);
      ExitProcess(0);
    }
  }
  else
  {
    if (!hMainWnd || !hAkelProcess)
    {
      MessageBoxW(NULL, GetLangStringW(wLangSystem, STRID_WINDOW_NOT_FOUND), L"Scripts Plugin", MB_OK|MB_ICONEXCLAMATION);
      ExitProcess(0);
    }
    if (dwArchitectVersion == (DWORD)-1)
    {
      MessageBoxW(NULL, GetLangStringW(wLangSystem, STRID_ARCHITECTURE_MISMATCH), L"Scripts plugin", MB_OK|MB_ICONEXCLAMATION);
      ExitProcess(0);
    }
  }
}

void GetMainInfo()
{
  if (bOldWindows)
  {
    if (!(hMainWnd=FindWindowA(APP_MAIN_CLASS_3A, NULL)))
    {
      hMainWnd=FindWindowA(APP_MAIN_CLASS_4A, NULL);
    }
    if (hMainWnd)
    {
      GetWindowThreadProcessId(hMainWnd, &dwAkelProcessId);
      hAkelProcess=OpenProcess(PROCESS_ALL_ACCESS, FALSE, dwAkelProcessId);
    }
  }
  else
  {
    if (!(hMainWnd=FindWindowW(APP_MAIN_CLASS_3W, NULL)))
    {
      hMainWnd=FindWindowW(APP_MAIN_CLASS_4W, NULL);
    }
    if (hMainWnd)
    {
      GetWindowThreadProcessId(hMainWnd, &dwAkelProcessId);
      hAkelProcess=OpenProcess(PROCESS_ALL_ACCESS, FALSE, dwAkelProcessId);
    }
  }
  wLangSystem=PRIMARYLANGID(GetUserDefaultLangID());

  //Get main info
  if (hMainWnd)
  {
    void *pAkelBuffer;

    if (!dwArchitectVersion)
    {
      dwArchitectVersion=SendMessage(hMainWnd, AKD_PROGRAMARCHITECTURE, 0, 0);

      if (LOBYTE(dwArchitectVersion) != LOBYTE(AKELDLL) ||
          HIBYTE(dwArchitectVersion) != HIBYTE(AKELDLL))
      {
        dwArchitectVersion=(DWORD)-1;
      }
    }
    if (dwArchitectVersion != (DWORD)-1)
    {
      if (pAkelBuffer=(void *)SendMessage(hMainWnd, AKD_GLOBALALLOC, GPTR, sizeof(PLUGINDATA)))
      {
        SendMessage(hMainWnd, AKD_GETMAININFO, 0, (LPARAM)pAkelBuffer);
        ReadProcessMemory(hAkelProcess, pAkelBuffer, &pdMainInfo, sizeof(PLUGINDATA), NULL);
        SendMessage(hMainWnd, AKD_GLOBALFREE, (WPARAM)pAkelBuffer, 0);
      }
      wLangSystem=PRIMARYLANGID(pdMainInfo.wLangSystem);
      hAkelInstance=pdMainInfo.hInstanceEXE;
    }
  }
}

BOOL GetEditInfo(HWND hWnd, EDITINFO *ei)
{
  void *pAkelBuffer;
  BOOL bResult=FALSE;

  if (pAkelBuffer=(void *)SendMessage(hMainWnd, AKD_GLOBALALLOC, GPTR, sizeof(EDITINFO)))
  {
    if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)hWnd, (LPARAM)pAkelBuffer))
    {
      if (ReadProcessMemory(hAkelProcess, pAkelBuffer, ei, sizeof(EDITINFO), NULL))
      {
        bResult=TRUE;
      }
    }
    SendMessage(hMainWnd, AKD_GLOBALFREE, (WPARAM)pAkelBuffer, 0);
  }
  return bResult;
}

HRESULT STDMETHODCALLTYPE Document_GetMainWnd(IDocument *this, int *hWnd)
{
  IsAkelPadReady();
  {
    if (!hMainWnd)
      GetMainInfo();
    *hWnd=(int)hMainWnd;
  }
  return NOERROR;
}

HRESULT STDMETHODCALLTYPE Document_GetAkelDir(IDocument *this, BSTR *wpDir)
{
  HRESULT hr=NOERROR;

  IsAkelPadReady();
  {
    char szDir[MAX_PATH];
    wchar_t wszDir[MAX_PATH];

    if (bOldWindows)
    {
      ReadProcessMemory(hAkelProcess, pdMainInfo.pAkelDir, szDir, MAX_PATH, NULL);
      MultiByteToWideChar(CP_ACP, 0, szDir, -1, wszDir, MAX_PATH);
      if (!(*wpDir=SysAllocString(wszDir)))
        hr=E_OUTOFMEMORY;
    }
    else
    {
      ReadProcessMemory(hAkelProcess, pdMainInfo.pAkelDir, wszDir, MAX_PATH * sizeof(wchar_t), NULL);
      if (!(*wpDir=SysAllocString(wszDir)))
        hr=E_OUTOFMEMORY;
    }
  }
  return hr;
}

HRESULT STDMETHODCALLTYPE Document_GetInstanceDll(IDocument *this, int *hInstance)
{
  *hInstance=(int)hInstanceDLL;
  return NOERROR;
}

HRESULT STDMETHODCALLTYPE Document_IsOldWindows(IDocument *this, BOOL *bIsOld)
{
  *bIsOld=bOldWindows;
  return NOERROR;
}

HRESULT STDMETHODCALLTYPE Document_IsOldRichEdit(IDocument *this, BOOL *bIsOld)
{
  IsAkelPadReady();
  {
    *bIsOld=pdMainInfo.bOldRichEdit;
  }
  return NOERROR;
}

HRESULT STDMETHODCALLTYPE Document_IsOldComctl32(IDocument *this, BOOL *bIsOld)
{
  IsAkelPadReady();
  {
    *bIsOld=pdMainInfo.bOldComctl32;
  }
  return NOERROR;
}

HRESULT STDMETHODCALLTYPE Document_IsAkelEdit(IDocument *this, BOOL *bIsAkelEdit)
{
  IsAkelPadReady();
  {
    *bIsAkelEdit=pdMainInfo.bAkelEdit;
  }
  return NOERROR;
}

HRESULT STDMETHODCALLTYPE Document_IsMDI(IDocument *this, BOOL *bIsMDI)
{
  IsAkelPadReady();
  {
    *bIsMDI=pdMainInfo.bMDI;
  }
  return NOERROR;
}

HRESULT STDMETHODCALLTYPE Document_GetEditWnd(IDocument *this, int *hWnd)
{
  IsAkelPadReady();
  {
    EDITINFO ei;

    if (GetEditInfo(NULL, &ei))
      *hWnd=(int)ei.hWndEdit;
    else
      *hWnd=0;
  }
  return NOERROR;
}

HRESULT STDMETHODCALLTYPE Document_GetEditFile(IDocument *this, HWND hWnd, BSTR *wpFile)
{
  HRESULT hr=NOERROR;

  IsAkelPadReady();
  {
    char szFile[MAX_PATH];
    wchar_t wszFile[MAX_PATH];
    EDITINFO ei;

    *wpFile=NULL;
    GetEditInfo(hWnd, &ei);

    if (ei.hWndEdit)
    {
      if (bOldWindows)
      {
        if (ReadProcessMemory(hAkelProcess, ei.pFile, szFile, MAX_PATH, NULL))
        {
          MultiByteToWideChar(CP_ACP, 0, szFile, -1, wszFile, MAX_PATH);
          if (!(*wpFile=SysAllocString(wszFile)))
            hr=E_OUTOFMEMORY;
        }
      }
      else
      {
        if (ReadProcessMemory(hAkelProcess, ei.pFile, wszFile, MAX_PATH * sizeof(wchar_t), NULL))
        {
          if (!(*wpFile=SysAllocString(wszFile)))
            hr=E_OUTOFMEMORY;
        }
      }
    }
  }
  return hr;
}

HRESULT STDMETHODCALLTYPE Document_GetEditCodePage(IDocument *this, HWND hWnd, int *nCodePage)
{
  IsAkelPadReady();
  {
    EDITINFO ei;

    GetEditInfo(hWnd, &ei);
    *nCodePage=(int)ei.nCodePage;
  }
  return NOERROR;
}

HRESULT STDMETHODCALLTYPE Document_GetEditBOM(IDocument *this, HWND hWnd, BOOL *bBOM)
{
  IsAkelPadReady();
  {
    EDITINFO ei;

    GetEditInfo(hWnd, &ei);
    *bBOM=(int)ei.bBOM;
  }
  return NOERROR;
}

HRESULT STDMETHODCALLTYPE Document_GetEditNewLine(IDocument *this, HWND hWnd, int *nNewLine)
{
  IsAkelPadReady();
  {
    EDITINFO ei;

    GetEditInfo(hWnd, &ei);
    *nNewLine=(int)ei.nNewLine;
  }
  return NOERROR;
}

HRESULT STDMETHODCALLTYPE Document_GetEditModified(IDocument *this, HWND hWnd, BOOL *bModified)
{
  IsAkelPadReady();
  {
    EDITINFO ei;

    GetEditInfo(hWnd, &ei);
    *bModified=(int)ei.bModified;
  }
  return NOERROR;
}

HRESULT STDMETHODCALLTYPE Document_GetEditReadOnly(IDocument *this, HWND hWnd, BOOL *bReadOnly)
{
  IsAkelPadReady();
  {
    EDITINFO ei;

    GetEditInfo(hWnd, &ei);
    *bReadOnly=(int)ei.bReadOnly;
  }
  return NOERROR;
}

HRESULT STDMETHODCALLTYPE Document_SendMessage(IDocument *this, int hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, LRESULT *lResult)
{
  if (bOldWindows)
    *lResult=SendMessageA((HWND)hWnd, uMsg, wParam, lParam);
  else
    *lResult=SendMessageW((HWND)hWnd, uMsg, wParam, lParam);
  return NOERROR;
}

HRESULT STDMETHODCALLTYPE Document_MessageBox(IDocument *this, int hWnd, BSTR pText, BSTR pCaption, int nType, int *nResult)
{
  *nResult=MessageBoxW((HWND)hWnd, pText, pCaption, nType);
  return NOERROR;
}

HRESULT STDMETHODCALLTYPE Document_InputBox(IDocument *this, int hWnd, BSTR wpCaption, BSTR wpLabel, BSTR wpEdit, BSTR *wpResult)
{
  INPUTBOX db;

  *wpResult=NULL;

  db.wpCaption=wpCaption;
  db.wpLabel=wpLabel;
  db.wpEdit=wpEdit;
  db.wpResult=wpResult;
  db.hr=NOERROR;

  if (bOldWindows)
    DialogBoxParamA(hInstanceDLL, MAKEINTRESOURCEA(IDD_INPUTBOX), (HWND)hWnd, (DLGPROC)InputBoxProcA, (LPARAM)&db);
  else
    DialogBoxParamW(hInstanceDLL, MAKEINTRESOURCEW(IDD_INPUTBOX), (HWND)hWnd, (DLGPROC)InputBoxProcW, (LPARAM)&db);

  return db.hr;
}

BOOL CALLBACK InputBoxProcA(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static HWND hWndEditBox;
  static INPUTBOX *db;

  if (uMsg == WM_INITDIALOG)
  {
    char szLabel[MAX_PATH];
    char szCaption[MAX_PATH];
    char szEdit[MAX_PATH];

    db=(INPUTBOX *)lParam;

    SendMessage(hDlg, WM_SETICON, (WPARAM)ICON_BIG, (LPARAM)hIconDLL);
    hWndEditBox=GetDlgItem(hDlg, IDC_INPUTBOX_EDIT);

    SetDlgItemTextA(hDlg, IDOK, GetLangStringA(wLangSystem, STRID_OK));
    SetDlgItemTextA(hDlg, IDCANCEL, GetLangStringA(wLangSystem, STRID_CANCEL));

    WideCharToMultiByte(CP_ACP, 0, db->wpLabel, -1, szLabel, MAX_PATH, NULL, NULL);
    SetDlgItemTextA(hDlg, IDC_INPUTBOX_LABEL, szLabel);

    WideCharToMultiByte(CP_ACP, 0, db->wpCaption, -1, szCaption, MAX_PATH, NULL, NULL);
    SetWindowTextA(hDlg, szCaption);

    WideCharToMultiByte(CP_ACP, 0, db->wpEdit, -1, szEdit, MAX_PATH, NULL, NULL);
    SetWindowTextA(hWndEditBox, szEdit);
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDOK)
    {
      char *szText;
      wchar_t *wszText;
      int nTextLen;

      nTextLen=GetWindowTextLengthA(hWndEditBox);

      if (szText=(char *)GlobalAlloc(GPTR, nTextLen + 1))
      {
        if (GetWindowTextA(hWndEditBox, szText, nTextLen + 1))
        {
          if (wszText=(wchar_t *)GlobalAlloc(GPTR, nTextLen * sizeof(wchar_t) + 2))
          {
            MultiByteToWideChar(CP_ACP, 0, szText, nTextLen + 1, wszText, nTextLen + 1);
            if (!(*db->wpResult=SysAllocString(wszText)))
              db->hr=E_OUTOFMEMORY;

            GlobalFree((HGLOBAL)wszText);
          }
        }
        GlobalFree((HGLOBAL)szText);
      }
      EndDialog(hDlg, 0);
      return TRUE;
    }
    else if (LOWORD(wParam) == IDCANCEL)
    {
      EndDialog(hDlg, 0);
      return TRUE;
    }
  }
  return FALSE;
}

BOOL CALLBACK InputBoxProcW(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static HWND hWndEditBox;
  static INPUTBOX *db;

  if (uMsg == WM_INITDIALOG)
  {
    db=(INPUTBOX *)lParam;

    SendMessage(hDlg, WM_SETICON, (WPARAM)ICON_BIG, (LPARAM)hIconDLL);
    hWndEditBox=GetDlgItem(hDlg, IDC_INPUTBOX_EDIT);

    SetDlgItemTextW(hDlg, IDOK, GetLangStringW(wLangSystem, STRID_OK));
    SetDlgItemTextW(hDlg, IDCANCEL, GetLangStringW(wLangSystem, STRID_CANCEL));
    SetDlgItemTextW(hDlg, IDC_INPUTBOX_LABEL, db->wpLabel);
    SetWindowTextW(hDlg, db->wpCaption);
    SetWindowTextW(hWndEditBox, db->wpEdit);
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDOK)
    {
      wchar_t *wszText;
      int nTextLen;

      nTextLen=GetWindowTextLengthW(hWndEditBox);

      if (wszText=(wchar_t *)GlobalAlloc(GPTR, nTextLen * sizeof(wchar_t) + 2))
      {
        if (GetWindowTextW(hWndEditBox, wszText, nTextLen + 1))
        {
          if (!(*db->wpResult=SysAllocString(wszText)))
            db->hr=E_OUTOFMEMORY;
        }
        GlobalFree((HGLOBAL)wszText);
      }
      EndDialog(hDlg, 0);
      return TRUE;
    }
    else if (LOWORD(wParam) == IDCANCEL)
    {
      EndDialog(hDlg, 0);
      return TRUE;
    }
  }
  return FALSE;
}

HRESULT STDMETHODCALLTYPE Document_GetSelStart(IDocument *this, int *nSelStart)
{
  IsAkelPadReady();
  {
    EDITINFO ei;
    CHARRANGE cr={0};
    void *pAkelBuffer;

    *nSelStart=0;
    GetEditInfo(NULL, &ei);

    if (ei.hWndEdit)
    {
      if (pAkelBuffer=(void *)SendMessage(hMainWnd, AKD_GLOBALALLOC, GPTR, sizeof(CHARRANGE)))
      {
        SendMessage(ei.hWndEdit, EM_EXGETSEL, 0, (LPARAM)pAkelBuffer);
        ReadProcessMemory(hAkelProcess, pAkelBuffer, &cr, sizeof(CHARRANGE), NULL);
        SendMessage(hMainWnd, AKD_GLOBALFREE, (WPARAM)pAkelBuffer, 0);
      }
      *nSelStart=cr.cpMin;
    }
  }
  return NOERROR;
}

HRESULT STDMETHODCALLTYPE Document_GetSelEnd(IDocument *this, int *nSelEnd)
{
  IsAkelPadReady();
  {
    EDITINFO ei;
    CHARRANGE cr={0};
    void *pAkelBuffer;

    *nSelEnd=0;
    GetEditInfo(NULL, &ei);

    if (ei.hWndEdit)
    {
      if (pAkelBuffer=(void *)SendMessage(hMainWnd, AKD_GLOBALALLOC, GPTR, sizeof(CHARRANGE)))
      {
        SendMessage(ei.hWndEdit, EM_EXGETSEL, 0, (LPARAM)pAkelBuffer);
        ReadProcessMemory(hAkelProcess, pAkelBuffer, &cr, sizeof(CHARRANGE), NULL);
        SendMessage(hMainWnd, AKD_GLOBALFREE, (WPARAM)pAkelBuffer, 0);
      }
      *nSelEnd=cr.cpMax;
    }
  }
  return NOERROR;
}

HRESULT STDMETHODCALLTYPE Document_SetSel(IDocument *this, int nSelStart, int nSelEnd)
{
  IsAkelPadReady();
  {
    EDITINFO ei;
    CHARRANGE cr;
    void *pAkelBuffer;

    GetEditInfo(NULL, &ei);

    if (ei.hWndEdit)
    {
      cr.cpMin=nSelStart;
      cr.cpMax=nSelEnd;

      if (pAkelBuffer=(void *)SendMessage(hMainWnd, AKD_GLOBALALLOC, GPTR, sizeof(CHARRANGE)))
      {
        if (WriteProcessMemory(hAkelProcess, pAkelBuffer, &cr, sizeof(CHARRANGE), NULL))
          SendMessage(ei.hWndEdit, EM_EXSETSEL, 0, (LPARAM)pAkelBuffer);
        SendMessage(hMainWnd, AKD_GLOBALFREE, (WPARAM)pAkelBuffer, 0);
      }
    }
  }
  return NOERROR;
}

HRESULT STDMETHODCALLTYPE Document_GetSelText(IDocument *this, int nNewLine, BSTR *wpText)
{
  HRESULT hr=NOERROR;

  IsAkelPadReady();
  {
    EDITINFO ei;
    BSTR wszText;
    int nTextLen;

    *wpText=NULL;
    GetEditInfo(NULL, &ei);

    if (ei.hWndEdit)
    {
      if (pdMainInfo.bAkelEdit)
      {
        EXGETTEXTRANGE tr={0};
        EXGETTEXTRANGE *pAkelTextRange;

        if (nNewLine == 2)
          tr.nNewLine=AELB_N;
        else if (nNewLine == 3)
          tr.nNewLine=AELB_RN;
        else
          tr.nNewLine=AELB_R;

        if (pAkelTextRange=(EXGETTEXTRANGE *)SendMessage(hMainWnd, AKD_GLOBALALLOC, GPTR, sizeof(EXGETTEXTRANGE)))
        {
          if (WriteProcessMemory(hAkelProcess, pAkelTextRange, &tr, sizeof(EXGETTEXTRANGE), NULL))
          {
            SendMessage(ei.hWndEdit, AEM_GETSEL, (WPARAM)NULL, (LPARAM)&pAkelTextRange->cr);

            if (nTextLen=SendMessage(hMainWnd, AKD_EXGETTEXTRANGEW, (WPARAM)ei.hWndEdit, (LPARAM)pAkelTextRange))
            {
              if (ReadProcessMemory(hAkelProcess, pAkelTextRange, &tr, sizeof(EXGETTEXTRANGE), NULL))
              {
                if (wszText=(BSTR)GlobalAlloc(GPTR, nTextLen * sizeof(wchar_t) + 2))
                {
                  if (ReadProcessMemory(hAkelProcess, tr.pText, wszText, nTextLen * sizeof(wchar_t) + 2, NULL))
                  {
                    if (!(*wpText=SysAllocStringLen(wszText, nTextLen)))
                      hr=E_OUTOFMEMORY;
                  }
                  GlobalFree((HGLOBAL)wszText);
                }
                else hr=E_OUTOFMEMORY;

                SendMessage(hMainWnd, AKD_FREETEXT, 0, (LPARAM)tr.pText);
              }
            }
          }
          SendMessage(hMainWnd, AKD_GLOBALFREE, (WPARAM)pAkelTextRange, 0);
        }
      }
      else
      {
        CHARRANGE cr;
        void *pAkelBuffer;

        if (pAkelBuffer=(void *)SendMessage(hMainWnd, AKD_GLOBALALLOC, GPTR, sizeof(CHARRANGE)))
        {
          SendMessage(ei.hWndEdit, EM_EXGETSEL, 0, (LPARAM)pAkelBuffer);
          ReadProcessMemory(hAkelProcess, pAkelBuffer, &cr, sizeof(CHARRANGE), NULL);
          SendMessage(hMainWnd, AKD_GLOBALFREE, (WPARAM)pAkelBuffer, 0);
        }

        if (cr.cpMin < cr.cpMax)
        {
          nTextLen=(cr.cpMax - cr.cpMin);

          if (pAkelBuffer=(void *)SendMessage(hMainWnd, AKD_GETSELTEXTW, (WPARAM)ei.hWndEdit, 0))
          {
            if (wszText=(BSTR)GlobalAlloc(GPTR, nTextLen * sizeof(wchar_t) + 2))
            {
              if (ReadProcessMemory(hAkelProcess, pAkelBuffer, wszText, nTextLen * sizeof(wchar_t) + 2, NULL))
              {
                if (!(*wpText=SysAllocStringLen(wszText, nTextLen)))
                  hr=E_OUTOFMEMORY;
              }
              GlobalFree((HGLOBAL)wszText);
            }
            else hr=E_OUTOFMEMORY;

            SendMessage(hMainWnd, AKD_FREETEXT, 0, (LPARAM)pAkelBuffer);
          }
        }
      }
    }
  }
  return hr;
}

HRESULT STDMETHODCALLTYPE Document_ReplaceSel(IDocument *this, BSTR wpText)
{
  IsAkelPadReady();
  {
    EDITINFO ei;
    void *pAkelBuffer;
    int nTextLen;

    GetEditInfo(NULL, &ei);

    if (ei.hWndEdit)
    {
      nTextLen=SysStringLen(wpText) * sizeof(wchar_t) + 2;

      if (pAkelBuffer=(void *)SendMessage(hMainWnd, AKD_GLOBALALLOC, GPTR, nTextLen))
      {
        if (WriteProcessMemory(hAkelProcess, pAkelBuffer, wpText, nTextLen, NULL))
          SendMessage(hMainWnd, AKD_REPLACESELW, (WPARAM)ei.hWndEdit, (LPARAM)pAkelBuffer);
        SendMessage(hMainWnd, AKD_GLOBALFREE, (WPARAM)pAkelBuffer, 0);
      }
    }
  }
  return NOERROR;
}

HRESULT STDMETHODCALLTYPE Document_TextFind(IDocument *this, HWND hWnd, BSTR wpFindIt, DWORD dwFlags, int *nResult)
{
  IsAkelPadReady();
  {
    *nResult=0;

    if (bOldWindows)
    {
      TEXTFINDA tf;
      void *pAkelFindIt;
      void *pAkelBuffer;
      int nFindItLenA;
      int nFindItLenW;
      char *pFindIt;

      nFindItLenW=SysStringLen(wpFindIt) + 1;
      nFindItLenA=WideCharToMultiByte(CP_ACP, 0, wpFindIt, nFindItLenW, NULL, 0, NULL, NULL);

      if (pFindIt=(char *)GlobalAlloc(GPTR, nFindItLenA))
      {
        WideCharToMultiByte(CP_ACP, 0, wpFindIt, nFindItLenW, pFindIt, nFindItLenA, NULL, NULL);

        if (pAkelFindIt=(void *)SendMessage(hMainWnd, AKD_GLOBALALLOC, GPTR, nFindItLenA))
        {
          if (WriteProcessMemory(hAkelProcess, pAkelFindIt, pFindIt, nFindItLenA, NULL))
          {
            tf.dwFlags=dwFlags;
            tf.pFindIt=pAkelFindIt;

            if (pAkelBuffer=(void *)SendMessage(hMainWnd, AKD_GLOBALALLOC, GPTR, sizeof(TEXTFINDA)))
            {
              if (WriteProcessMemory(hAkelProcess, pAkelBuffer, &tf, sizeof(TEXTFINDA), NULL))
                *nResult=SendMessage(hMainWnd, AKD_TEXTFIND, (WPARAM)hWnd, (LPARAM)pAkelBuffer);
              SendMessage(hMainWnd, AKD_GLOBALFREE, (WPARAM)pAkelBuffer, 0);
            }
          }
          SendMessage(hMainWnd, AKD_GLOBALFREE, (WPARAM)pAkelFindIt, 0);
        }
        GlobalFree((HGLOBAL)pFindIt);
      }
    }
    else
    {
      TEXTFINDW tf;
      void *pAkelFindIt;
      void *pAkelBuffer;
      int nFindItBytes;

      nFindItBytes=SysStringLen(wpFindIt) * sizeof(wchar_t) + 2;

      if (pAkelFindIt=(void *)SendMessage(hMainWnd, AKD_GLOBALALLOC, GPTR, nFindItBytes))
      {
        if (WriteProcessMemory(hAkelProcess, pAkelFindIt, wpFindIt, nFindItBytes, NULL))
        {
          tf.dwFlags=dwFlags;
          tf.pFindIt=pAkelFindIt;

          if (pAkelBuffer=(void *)SendMessage(hMainWnd, AKD_GLOBALALLOC, GPTR, sizeof(TEXTFINDW)))
          {
            if (WriteProcessMemory(hAkelProcess, pAkelBuffer, &tf, sizeof(TEXTFINDW), NULL))
              *nResult=SendMessage(hMainWnd, AKD_TEXTFIND, (WPARAM)hWnd, (LPARAM)pAkelBuffer);
            SendMessage(hMainWnd, AKD_GLOBALFREE, (WPARAM)pAkelBuffer, 0);
          }
        }
        SendMessage(hMainWnd, AKD_GLOBALFREE, (WPARAM)pAkelFindIt, 0);
      }
    }
  }
  return NOERROR;
}

HRESULT STDMETHODCALLTYPE Document_TextReplace(IDocument *this, HWND hWnd, BSTR wpFindIt, BSTR wpReplaceWith, DWORD dwFlags, BOOL bAll, int *nResult)
{
  IsAkelPadReady();
  {
    *nResult=0;

    if (bOldWindows)
    {
      TEXTREPLACEA tr;
      void *pAkelFindIt;
      void *pAkelReplaceWith;
      void *pAkelBuffer;
      int nFindItLenA;
      int nFindItLenW;
      int nReplaceWithLenA;
      int nReplaceWithLenW;
      char *pFindIt;
      char *pReplaceWith;

      nFindItLenW=SysStringLen(wpFindIt) + 1;
      nFindItLenA=WideCharToMultiByte(CP_ACP, 0, wpFindIt, nFindItLenW, NULL, 0, NULL, NULL);

      if (pFindIt=(char *)GlobalAlloc(GPTR, nFindItLenA))
      {
        WideCharToMultiByte(CP_ACP, 0, wpFindIt, nFindItLenW, pFindIt, nFindItLenA, NULL, NULL);

        if (pAkelFindIt=(void *)SendMessage(hMainWnd, AKD_GLOBALALLOC, GPTR, nFindItLenA))
        {
          if (WriteProcessMemory(hAkelProcess, pAkelFindIt, pFindIt, nFindItLenA, NULL))
          {
            nReplaceWithLenW=SysStringLen(wpReplaceWith) + 1;
            nReplaceWithLenA=WideCharToMultiByte(CP_ACP, 0, wpReplaceWith, nReplaceWithLenW, NULL, 0, NULL, NULL);

            if (pReplaceWith=(char *)GlobalAlloc(GPTR, nReplaceWithLenA))
            {
              WideCharToMultiByte(CP_ACP, 0, wpReplaceWith, nReplaceWithLenW, pReplaceWith, nReplaceWithLenA, NULL, NULL);

              if (pAkelReplaceWith=(void *)SendMessage(hMainWnd, AKD_GLOBALALLOC, GPTR, nReplaceWithLenA))
              {
                if (WriteProcessMemory(hAkelProcess, pAkelReplaceWith, pReplaceWith, nReplaceWithLenA, NULL))
                {
                  tr.dwFlags=dwFlags;
                  tr.pFindIt=pAkelFindIt;
                  tr.pReplaceWith=pAkelReplaceWith;
                  tr.bAll=bAll;

                  if (pAkelBuffer=(void *)SendMessage(hMainWnd, AKD_GLOBALALLOC, GPTR, sizeof(TEXTREPLACEA)))
                  {
                    if (WriteProcessMemory(hAkelProcess, pAkelBuffer, &tr, sizeof(TEXTREPLACEA), NULL))
                      *nResult=SendMessage(hMainWnd, AKD_TEXTREPLACE, (WPARAM)hWnd, (LPARAM)pAkelBuffer);
                    SendMessage(hMainWnd, AKD_GLOBALFREE, (WPARAM)pAkelBuffer, 0);
                  }
                }
                SendMessage(hMainWnd, AKD_GLOBALFREE, (WPARAM)pAkelReplaceWith, 0);
              }
              GlobalFree((HGLOBAL)pReplaceWith);
            }
          }
          SendMessage(hMainWnd, AKD_GLOBALFREE, (WPARAM)pAkelFindIt, 0);
        }
        GlobalFree((HGLOBAL)pFindIt);
      }
    }
    else
    {
      TEXTREPLACEW tr;
      void *pAkelFindIt;
      void *pAkelReplaceWith;
      void *pAkelBuffer;
      int nFindItBytes;
      int nReplaceWithBytes;

      nFindItBytes=SysStringLen(wpFindIt) * sizeof(wchar_t) + 2;

      if (pAkelFindIt=(void *)SendMessage(hMainWnd, AKD_GLOBALALLOC, GPTR, nFindItBytes))
      {
        if (WriteProcessMemory(hAkelProcess, pAkelFindIt, wpFindIt, nFindItBytes, NULL))
        {
          nReplaceWithBytes=SysStringLen(wpReplaceWith) * sizeof(wchar_t) + 2;

          if (pAkelReplaceWith=(void *)SendMessage(hMainWnd, AKD_GLOBALALLOC, GPTR, nReplaceWithBytes))
          {
            if (WriteProcessMemory(hAkelProcess, pAkelReplaceWith, wpReplaceWith, nReplaceWithBytes, NULL))
            {
              tr.dwFlags=dwFlags;
              tr.pFindIt=pAkelFindIt;
              tr.pReplaceWith=pAkelReplaceWith;
              tr.bAll=bAll;

              if (pAkelBuffer=(void *)SendMessage(hMainWnd, AKD_GLOBALALLOC, GPTR, sizeof(TEXTREPLACEW)))
              {
                if (WriteProcessMemory(hAkelProcess, pAkelBuffer, &tr, sizeof(TEXTREPLACEW), NULL))
                  *nResult=SendMessage(hMainWnd, AKD_TEXTREPLACE, (WPARAM)hWnd, (LPARAM)pAkelBuffer);
                SendMessage(hMainWnd, AKD_GLOBALFREE, (WPARAM)pAkelBuffer, 0);
              }
            }
            SendMessage(hMainWnd, AKD_GLOBALFREE, (WPARAM)pAkelReplaceWith, 0);
          }
        }
        SendMessage(hMainWnd, AKD_GLOBALFREE, (WPARAM)pAkelFindIt, 0);
      }
    }
  }
  return NOERROR;
}

HRESULT STDMETHODCALLTYPE Document_GetClipboardText(IDocument *this, BOOL bAnsi, BSTR *wpText)
{
  HRESULT hr=NOERROR;
  HGLOBAL hData;
  LPVOID pData;

  *wpText=NULL;

  if (OpenClipboard(NULL))
  {
    if (!bAnsi && (hData=GetClipboardData(CF_UNICODETEXT)))
    {
      if (pData=GlobalLock(hData))
      {
        if (!(*wpText=SysAllocString((wchar_t *)pData)))
          hr=E_OUTOFMEMORY;
        GlobalUnlock(hData);
      }
    }
    else if (hData=GetClipboardData(CF_TEXT))
    {
      if (pData=GlobalLock(hData))
      {
        wchar_t *wszBuffer;
        int nUnicodeLen;

        nUnicodeLen=MultiByteToWideChar(CP_ACP, 0, (char *)pData, -1, NULL, 0);

        if (wszBuffer=(wchar_t *)GlobalAlloc(GPTR, nUnicodeLen * sizeof(wchar_t)))
        {
          MultiByteToWideChar(CP_ACP, 0, (char *)pData, -1, wszBuffer, nUnicodeLen);
          if (!(*wpText=SysAllocString(wszBuffer)))
            hr=E_OUTOFMEMORY;
          GlobalFree((HGLOBAL)wszBuffer);
        }
        else hr=E_OUTOFMEMORY;

        GlobalUnlock(hData);
      }
    }
    CloseClipboard();
  }
  return hr;
}

HRESULT STDMETHODCALLTYPE Document_SetClipboardText(IDocument *this, BSTR wpText)
{
  HRESULT hr=NOERROR;
  HGLOBAL hDataA=NULL;
  HGLOBAL hDataW=NULL;
  LPVOID pData;
  int nUnicodeLen;
  int nAnsiLen;

  if (OpenClipboard(NULL))
  {
    //Unicode
    nUnicodeLen=SysStringLen(wpText) + 1;

    if (hDataW=GlobalAlloc(GMEM_MOVEABLE, nUnicodeLen * sizeof(wchar_t)))
    {
      if (pData=GlobalLock(hDataW))
      {
        memcpy(pData, wpText, nUnicodeLen * sizeof(wchar_t));
        GlobalUnlock(hDataW);
      }
    }

    //ANSI
    nAnsiLen=WideCharToMultiByte(CP_ACP, 0, wpText, nUnicodeLen, NULL, 0, NULL, NULL);

    if (hDataA=GlobalAlloc(GMEM_MOVEABLE, nAnsiLen))
    {
      if (pData=GlobalLock(hDataA))
      {
        WideCharToMultiByte(CP_ACP, 0, wpText, nUnicodeLen, (char *)pData, nAnsiLen, NULL, NULL);
        GlobalUnlock(hDataA);
      }
    }
    EmptyClipboard();
    if (hDataW) SetClipboardData(CF_UNICODETEXT, hDataW);
    if (hDataA) SetClipboardData(CF_TEXT, hDataA);
    CloseClipboard();
  }
  return hr;
}

HRESULT STDMETHODCALLTYPE Document_Call(IDocument *this, BSTR wpFunction, SAFEARRAY **psa, LRESULT *lResult)
{
  HRESULT hr=NOERROR;

  IsAkelPadReady();
  {
    HSTACK hStringsStack={0};
    POINTER *lpElement;
    VARIANT *vtParameter;
    unsigned char *lpData;
    unsigned char *lpStructure=NULL;
    DWORD dwElement;
    DWORD dwElementSum;
    DWORD dwStructSize;
    DWORD dwParameter;
    DWORD dwOffset;
    void *pAkelString;

    *lResult=UD_FAILED;
    lpData=(unsigned char *)((*psa)->pvData);
    dwElementSum=(*psa)->rgsabound[0].cElements;
    dwStructSize=(dwElementSum + 1) * sizeof(DWORD);
    dwOffset=0;

    if (dwElementSum)
    {
      if (lpStructure=(unsigned char *)GlobalAlloc(GPTR, dwStructSize))
      {
        //Set structure size in first parameter
        *(DWORD *)lpStructure=dwStructSize;
        dwOffset+=sizeof(DWORD);

        for (dwElement=0; dwElement < dwElementSum; ++dwElement)
        {
          vtParameter=(VARIANT *)(lpData + dwElement * sizeof(VARIANT));
          dwParameter=0;

          if (vtParameter->vt == VT_BOOL)
            dwParameter=vtParameter->boolVal;
          else if (vtParameter->vt == VT_I2)
            dwParameter=vtParameter->iVal;
          else if (vtParameter->vt == VT_I4)
            dwParameter=vtParameter->lVal;
          else if (vtParameter->vt == VT_R8)
            dwParameter=(DWORD)vtParameter->dblVal;
          else if (vtParameter->vt == VT_BSTR)
          {
            if (bOldWindows)
            {
              char *szString;
              int nStringLenA;
              int nStringLenW;

              nStringLenW=SysStringLen(vtParameter->bstrVal) + 1;
              nStringLenA=WideCharToMultiByte(CP_ACP, 0, vtParameter->bstrVal, nStringLenW, NULL, 0, NULL, NULL);

              if (szString=(char *)GlobalAlloc(GPTR, nStringLenA))
              {
                WideCharToMultiByte(CP_ACP, 0, vtParameter->bstrVal, nStringLenW, szString, nStringLenA, NULL, NULL);

                if (pAkelString=(void *)SendMessage(hMainWnd, AKD_GLOBALALLOC, GPTR, nStringLenA))
                {
                  if (WriteProcessMemory(hAkelProcess, pAkelString, szString, nStringLenA, NULL))
                  {
                    if (!StackInsertIndex((stack **)&hStringsStack.first, (stack **)&hStringsStack.last, (stack **)&lpElement, -1, sizeof(POINTER)))
                    {
                      lpElement->lpString=pAkelString;
                      dwParameter=(DWORD)pAkelString;
                    }
                  }
                }
                GlobalFree((HGLOBAL)szString);
              }
            }
            else
            {
              int nStringBytes;

              nStringBytes=SysStringLen(vtParameter->bstrVal) * sizeof(wchar_t) + 2;

              if (pAkelString=(void *)SendMessage(hMainWnd, AKD_GLOBALALLOC, GPTR, nStringBytes))
              {
                if (WriteProcessMemory(hAkelProcess, pAkelString, vtParameter->bstrVal, nStringBytes, NULL))
                {
                  if (!StackInsertIndex((stack **)&hStringsStack.first, (stack **)&hStringsStack.last, (stack **)&lpElement, -1, sizeof(POINTER)))
                  {
                    lpElement->lpString=pAkelString;
                    dwParameter=(DWORD)pAkelString;
                  }
                }
              }
            }
          }
          *(DWORD *)(lpStructure + dwOffset)=dwParameter;
          dwOffset+=sizeof(DWORD);
        }
      }
    }

    if (bOldWindows)
    {
      PLUGINCALLSENDA pcs;
      char szFunction[MAX_PATH];
      void *pAkelFunction;
      void *pAkelFunctionStruct=NULL;
      void *pAkelBuffer;
      int nFunctionBytes;

      nFunctionBytes=WideCharToMultiByte(CP_ACP, 0, wpFunction, -1, szFunction, MAX_PATH, NULL, NULL);

      if (pAkelFunction=(void *)SendMessage(hMainWnd, AKD_GLOBALALLOC, GPTR, nFunctionBytes))
      {
        if (WriteProcessMemory(hAkelProcess, pAkelFunction, szFunction, nFunctionBytes, NULL))
        {
          if (lpStructure)
          {
            if (pAkelFunctionStruct=(void *)SendMessage(hMainWnd, AKD_GLOBALALLOC, GPTR, dwStructSize))
              WriteProcessMemory(hAkelProcess, pAkelFunctionStruct, lpStructure, dwStructSize, NULL);
          }
          pcs.pFunction=pAkelFunction;
          pcs.bOnStart=FALSE;
          pcs.lParam=(LPARAM)pAkelFunctionStruct;
          pcs.lpbAutoLoad=NULL;

          if (pAkelBuffer=(void *)SendMessage(hMainWnd, AKD_GLOBALALLOC, GPTR, sizeof(PLUGINCALLSENDA)))
          {
            if (WriteProcessMemory(hAkelProcess, pAkelBuffer, &pcs, sizeof(PLUGINCALLSENDA), NULL))
              *lResult=SendMessage(hMainWnd, AKD_DLLCALL, 0, (LPARAM)pAkelBuffer);
            SendMessage(hMainWnd, AKD_GLOBALFREE, (WPARAM)pAkelBuffer, 0);
          }
          if (pAkelFunctionStruct) SendMessage(hMainWnd, AKD_GLOBALFREE, (WPARAM)pAkelFunctionStruct, 0);
        }
        SendMessage(hMainWnd, AKD_GLOBALFREE, (WPARAM)pAkelFunction, 0);
      }
    }
    else
    {
      PLUGINCALLSENDW pcs;
      void *pAkelFunction;
      void *pAkelFunctionStruct=NULL;
      void *pAkelBuffer;
      int nFunctionBytes;

      nFunctionBytes=SysStringLen(wpFunction) * sizeof(wchar_t) + 2;

      if (pAkelFunction=(void *)SendMessage(hMainWnd, AKD_GLOBALALLOC, GPTR, nFunctionBytes))
      {
        if (WriteProcessMemory(hAkelProcess, pAkelFunction, wpFunction, nFunctionBytes, NULL))
        {
          if (lpStructure)
          {
            if (pAkelFunctionStruct=(void *)SendMessage(hMainWnd, AKD_GLOBALALLOC, GPTR, dwStructSize))
              WriteProcessMemory(hAkelProcess, pAkelFunctionStruct, lpStructure, dwStructSize, NULL);
          }
          pcs.pFunction=pAkelFunction;
          pcs.bOnStart=FALSE;
          pcs.lParam=(LPARAM)pAkelFunctionStruct;
          pcs.lpbAutoLoad=NULL;

          if (pAkelBuffer=(void *)SendMessage(hMainWnd, AKD_GLOBALALLOC, GPTR, sizeof(PLUGINCALLSENDW)))
          {
            if (WriteProcessMemory(hAkelProcess, pAkelBuffer, &pcs, sizeof(PLUGINCALLSENDW), NULL))
              *lResult=SendMessage(hMainWnd, AKD_DLLCALL, 0, (LPARAM)pAkelBuffer);
            SendMessage(hMainWnd, AKD_GLOBALFREE, (WPARAM)pAkelBuffer, 0);
          }
          if (pAkelFunctionStruct) SendMessage(hMainWnd, AKD_GLOBALFREE, (WPARAM)pAkelFunctionStruct, 0);
        }
        SendMessage(hMainWnd, AKD_GLOBALFREE, (WPARAM)pAkelFunction, 0);
      }
    }

    //Free structure
    if (lpStructure) GlobalFree((HGLOBAL)lpStructure);

    //Free strings stack
    lpElement=(POINTER *)hStringsStack.first;

    while (lpElement)
    {
      if (lpElement->lpString)
        SendMessage(hMainWnd, AKD_GLOBALFREE, (WPARAM)lpElement->lpString, 0);

      lpElement=lpElement->next;
    }
    StackClear((stack **)&hStringsStack.first, (stack **)&hStringsStack.last);
  }
  return hr;
}

HRESULT STDMETHODCALLTYPE Document_IsPluginRunning(IDocument *this, BSTR wpFunction, BOOL *bRunning)
{
  IsAkelPadReady();
  {
    *bRunning=FALSE;

    if (bOldWindows)
    {
      PLUGINFUNCTIONA *pf;
      char szFunction[MAX_PATH];
      void *pAkelFunction;
      int nFunctionBytes;

      nFunctionBytes=WideCharToMultiByte(CP_ACP, 0, wpFunction, -1, szFunction, MAX_PATH, NULL, NULL);

      if (pAkelFunction=(void *)SendMessage(hMainWnd, AKD_GLOBALALLOC, GPTR, nFunctionBytes))
      {
        if (WriteProcessMemory(hAkelProcess, pAkelFunction, szFunction, nFunctionBytes, NULL))
        {
          if (pf=(PLUGINFUNCTIONA *)SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)pAkelFunction, 0))
          {
            ReadProcessMemory(hAkelProcess, &pf->bRunning, bRunning, sizeof(BOOL), NULL);
          }
        }
        SendMessage(hMainWnd, AKD_GLOBALFREE, (WPARAM)pAkelFunction, 0);
      }
    }
    else
    {
      PLUGINFUNCTIONW *pf;
      void *pAkelFunction;
      int nFunctionBytes;

      nFunctionBytes=SysStringLen(wpFunction) * sizeof(wchar_t) + 2;

      if (pAkelFunction=(void *)SendMessage(hMainWnd, AKD_GLOBALALLOC, GPTR, nFunctionBytes))
      {
        if (WriteProcessMemory(hAkelProcess, pAkelFunction, wpFunction, nFunctionBytes, NULL))
        {
          if (pf=(PLUGINFUNCTIONW *)SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)pAkelFunction, 0))
          {
            ReadProcessMemory(hAkelProcess, &pf->bRunning, bRunning, sizeof(BOOL), NULL);
          }
          SendMessage(hMainWnd, AKD_GLOBALFREE, (WPARAM)pAkelFunction, 0);
        }
      }
    }
  }
  return NOERROR;
}

HRESULT STDMETHODCALLTYPE Document_OpenFile(IDocument *this, BSTR wpFile, DWORD dwFlags, int nCodePage, BOOL bBOM, int *nResult)
{
  IsAkelPadReady();
  {
    if (bOldWindows)
    {
      OPENDOCUMENTA od;
      COPYDATASTRUCT cds;
      char szFile[MAX_PATH];

      WideCharToMultiByte(CP_ACP, 0, wpFile, -1, szFile, MAX_PATH, NULL, NULL);

      od.hWnd=NULL;
      lstrcpynA(od.szFile, szFile, MAX_PATH);
      od.szWorkDir[0]='\0';
      od.dwFlags=dwFlags;
      od.nCodePage=nCodePage;
      od.bBOM=bBOM;

      cds.dwData=CD_OPENDOCUMENT;
      cds.cbData=sizeof(OPENDOCUMENTA);
      cds.lpData=&od;
      *nResult=SendMessage(hMainWnd, WM_COPYDATA, (WPARAM)hMainWnd, (LPARAM)&cds);
    }
    else
    {
      OPENDOCUMENTW od;
      COPYDATASTRUCT cds;

      od.hWnd=NULL;
      lstrcpynW(od.szFile, wpFile, MAX_PATH);
      od.szWorkDir[0]='\0';
      od.dwFlags=dwFlags;
      od.nCodePage=nCodePage;
      od.bBOM=bBOM;

      cds.dwData=CD_OPENDOCUMENT;
      cds.cbData=sizeof(OPENDOCUMENTW);
      cds.lpData=&od;
      *nResult=SendMessage(hMainWnd, WM_COPYDATA, (WPARAM)hMainWnd, (LPARAM)&cds);
    }
  }
  return NOERROR;
}

HRESULT STDMETHODCALLTYPE Document_ReadFile(IDocument *this, BSTR wpFile, DWORD dwFlags, int nCodePage, BOOL bBOM, BSTR *wpText)
{
  HRESULT hr=NOERROR;

  IsAkelPadReady();
  {
    BSTR wszText;
    void *pAkelWindowBuffer;
    void *pAkelTextBuffer;
    int nTextLen;

    *wpText=NULL;

    if (bOldWindows)
    {
      CREATEWINDOWA cw;
      OPENDOCUMENTA od;
      COPYDATASTRUCT cds;
      HWND hWndMemEdit;
      int nLangOptions;

      lstrcpynA(cw.szClassName, "RichEdit20A", MAX_PATH);
      cw.szWindowName[0]='\0';
      cw.dwStyle=WS_CHILD|WS_HSCROLL|WS_VSCROLL|ES_LEFT|ES_MULTILINE|ES_DISABLENOSCROLL|ES_SUNKEN|ES_NOHIDESEL;
      cw.x=0;
      cw.y=0;
      cw.nWidth=0;
      cw.nHeight=0;
      cw.hWndParent=hMainWnd;
      cw.hMenu=(HMENU)0;
      cw.hInstance=hAkelInstance;
      cw.lpParam=NULL;

      if (pAkelWindowBuffer=(void *)SendMessage(hMainWnd, AKD_GLOBALALLOC, GPTR, sizeof(CREATEWINDOWA)))
      {
        if (WriteProcessMemory(hAkelProcess, pAkelWindowBuffer, &cw, sizeof(CREATEWINDOWA), NULL))
        {
          if (hWndMemEdit=(HWND)SendMessage(hMainWnd, AKD_CREATEWINDOW, 0, (LPARAM)pAkelWindowBuffer))
          {
            //Edit adjustment
            SendMessage(hWndMemEdit, EM_EXLIMITTEXT, 0, 0x0FFFFFFF);
            nLangOptions=SendMessage(hWndMemEdit, EM_GETLANGOPTIONS, 0, 0);
            nLangOptions&=~IMF_AUTOFONT;
            nLangOptions&=~IMF_AUTOFONTSIZEADJUST;
            SendMessage(hWndMemEdit, EM_SETLANGOPTIONS, 0, nLangOptions);
            SendMessage(hWndMemEdit, EM_SETTEXTMODE, TM_PLAINTEXT|TM_MULTILEVELUNDO|TM_MULTICODEPAGE, 0);

            //Open file
            od.hWnd=hWndMemEdit;
            WideCharToMultiByte(CP_ACP, 0, wpFile, -1, od.szFile, MAX_PATH, NULL, NULL);
            GetCurrentDirectoryA(MAX_PATH, od.szWorkDir);
            od.dwFlags=dwFlags;
            od.nCodePage=nCodePage;
            od.bBOM=bBOM;

            cds.dwData=CD_OPENDOCUMENT;
            cds.cbData=sizeof(OPENDOCUMENTA);
            cds.lpData=&od;
            SendMessage(hMainWnd, WM_COPYDATA, (WPARAM)hMainWnd, (LPARAM)&cds);

            //Select all text
            SendMessage(hWndMemEdit, EM_SETSEL, 0, -1);

            //Get text
            if (nTextLen=SendMessage(hMainWnd, AKD_GETTEXTLENGTH, (WPARAM)hWndMemEdit, 0))
            {
              if (pAkelTextBuffer=(void *)SendMessage(hMainWnd, AKD_GETSELTEXTW, (WPARAM)hWndMemEdit, 0))
              {
                if (wszText=(BSTR)GlobalAlloc(GPTR, nTextLen * sizeof(wchar_t) + 2))
                {
                  if (ReadProcessMemory(hAkelProcess, pAkelTextBuffer, wszText, nTextLen * sizeof(wchar_t) + 2, NULL))
                  {
                    SendMessage(hMainWnd, AKD_FREETEXT, 0, (LPARAM)pAkelTextBuffer);
                    pAkelTextBuffer=NULL;

                    if (!(*wpText=SysAllocString(wszText)))
                      hr=E_OUTOFMEMORY;
                  }
                  GlobalFree((HGLOBAL)wszText);
                }
                else hr=E_OUTOFMEMORY;

                if (pAkelTextBuffer) SendMessage(hMainWnd, AKD_FREETEXT, 0, (LPARAM)pAkelTextBuffer);
              }
            }

            //Destroy edit window
            SendMessage(hWndMemEdit, WM_CLOSE, 0, 0);
          }
        }
        SendMessage(hMainWnd, AKD_GLOBALFREE, (WPARAM)pAkelWindowBuffer, 0);
      }
    }
    else
    {
      CREATEWINDOWW cw;
      OPENDOCUMENTW od;
      COPYDATASTRUCT cds;
      HWND hWndMemEdit;
      int nLangOptions;

      lstrcpynW(cw.szClassName, L"RichEdit20W", MAX_PATH);
      cw.szWindowName[0]='\0';
      cw.dwStyle=WS_CHILD|WS_HSCROLL|WS_VSCROLL|ES_LEFT|ES_MULTILINE|ES_DISABLENOSCROLL|ES_SUNKEN|ES_NOHIDESEL;
      cw.x=0;
      cw.y=0;
      cw.nWidth=0;
      cw.nHeight=0;
      cw.hWndParent=hMainWnd;
      cw.hMenu=(HMENU)0;
      cw.hInstance=hAkelInstance;
      cw.lpParam=NULL;

      if (pAkelWindowBuffer=(void *)SendMessage(hMainWnd, AKD_GLOBALALLOC, GPTR, sizeof(CREATEWINDOWW)))
      {
        if (WriteProcessMemory(hAkelProcess, pAkelWindowBuffer, &cw, sizeof(CREATEWINDOWW), NULL))
        {
          if (hWndMemEdit=(HWND)SendMessage(hMainWnd, AKD_CREATEWINDOW, 0, (LPARAM)pAkelWindowBuffer))
          {
            //Edit adjustment
            SendMessage(hWndMemEdit, EM_EXLIMITTEXT, 0, 0x0FFFFFFF);
            nLangOptions=SendMessage(hWndMemEdit, EM_GETLANGOPTIONS, 0, 0);
            nLangOptions&=~IMF_AUTOFONT;
            nLangOptions&=~IMF_AUTOFONTSIZEADJUST;
            SendMessage(hWndMemEdit, EM_SETLANGOPTIONS, 0, nLangOptions);
            SendMessage(hWndMemEdit, EM_SETTEXTMODE, TM_PLAINTEXT|TM_MULTILEVELUNDO|TM_MULTICODEPAGE, 0);

            //Open file
            od.hWnd=hWndMemEdit;
            lstrcpynW(od.szFile, wpFile, MAX_PATH);
            GetCurrentDirectoryW(MAX_PATH, od.szWorkDir);
            od.dwFlags=dwFlags;
            od.nCodePage=nCodePage;
            od.bBOM=bBOM;

            cds.dwData=CD_OPENDOCUMENT;
            cds.cbData=sizeof(OPENDOCUMENTW);
            cds.lpData=&od;
            SendMessage(hMainWnd, WM_COPYDATA, (WPARAM)hMainWnd, (LPARAM)&cds);

            //Select all text
            SendMessage(hWndMemEdit, EM_SETSEL, 0, -1);

            //Get text
            if (nTextLen=SendMessage(hMainWnd, AKD_GETTEXTLENGTH, (WPARAM)hWndMemEdit, 0))
            {
              if (pAkelTextBuffer=(void *)SendMessage(hMainWnd, AKD_GETSELTEXTW, (WPARAM)hWndMemEdit, 0))
              {
                if (wszText=(BSTR)GlobalAlloc(GPTR, nTextLen * sizeof(wchar_t) + 2))
                {
                  if (ReadProcessMemory(hAkelProcess, pAkelTextBuffer, wszText, nTextLen * sizeof(wchar_t) + 2, NULL))
                  {
                    SendMessage(hMainWnd, AKD_FREETEXT, 0, (LPARAM)pAkelTextBuffer);
                    pAkelTextBuffer=NULL;

                    if (!(*wpText=SysAllocString(wszText)))
                      hr=E_OUTOFMEMORY;
                  }
                  GlobalFree((HGLOBAL)wszText);
                }
                else hr=E_OUTOFMEMORY;

                if (pAkelTextBuffer) SendMessage(hMainWnd, AKD_FREETEXT, 0, (LPARAM)pAkelTextBuffer);
              }
            }

            //Destroy edit window
            SendMessage(hWndMemEdit, WM_CLOSE, 0, 0);
          }
        }
        SendMessage(hMainWnd, AKD_GLOBALFREE, (WPARAM)pAkelWindowBuffer, 0);
      }
    }
  }
  return hr;
}

HRESULT STDMETHODCALLTYPE Document_SaveFile(IDocument *this, HWND hWnd, BSTR wpFile, int nCodePage, BOOL bBOM, DWORD dwFlags, int *nResult)
{
  IsAkelPadReady();
  {
    *nResult=0;

    if (bOldWindows)
    {
      SAVEDOCUMENTA sd;
      void *pAkelBuffer;

      WideCharToMultiByte(CP_ACP, 0, wpFile, -1, sd.szFile, MAX_PATH, NULL, NULL);
      sd.nCodePage=nCodePage;
      sd.bBOM=bBOM;
      sd.dwFlags=dwFlags;

      if (pAkelBuffer=(void *)SendMessage(hMainWnd, AKD_GLOBALALLOC, GPTR, sizeof(SAVEDOCUMENTA)))
      {
        if (WriteProcessMemory(hAkelProcess, pAkelBuffer, &sd, sizeof(SAVEDOCUMENTA), NULL))
          *nResult=SendMessage(hMainWnd, AKD_SAVEDOCUMENT, (WPARAM)hWnd, (LPARAM)pAkelBuffer);
        SendMessage(hMainWnd, AKD_GLOBALFREE, (WPARAM)pAkelBuffer, 0);
      }
    }
    else
    {
      SAVEDOCUMENTW sd;
      void *pAkelBuffer;

      lstrcpynW(sd.szFile, wpFile, MAX_PATH);
      sd.nCodePage=nCodePage;
      sd.bBOM=bBOM;
      sd.dwFlags=dwFlags;

      if (pAkelBuffer=(void *)SendMessage(hMainWnd, AKD_GLOBALALLOC, GPTR, sizeof(SAVEDOCUMENTW)))
      {
        if (WriteProcessMemory(hAkelProcess, pAkelBuffer, &sd, sizeof(SAVEDOCUMENTW), NULL))
          *nResult=SendMessage(hMainWnd, AKD_SAVEDOCUMENT, (WPARAM)hWnd, (LPARAM)pAkelBuffer);
        SendMessage(hMainWnd, AKD_GLOBALFREE, (WPARAM)pAkelBuffer, 0);
      }
    }
  }
  return NOERROR;
}

HRESULT STDMETHODCALLTYPE Document_SystemFunction(IDocument *this, IDispatch **objFunction)
{
  IRealSystemFunction *lpFunction;

  if (!objFunction) return E_POINTER;

  if ((lpFunction=(IRealSystemFunction *)GlobalAlloc(GPTR, sizeof(IRealSystemFunction))))
  {
    lpFunction->lpVtbl=(ISystemFunctionVtbl *)&MyISystemFunctionVtbl;
    lpFunction->dwCount=1;

    InterlockedIncrement(&g_nObjs);
    *objFunction=(IDispatch *)lpFunction;
  }
  else return E_OUTOFMEMORY;

  return NOERROR;
}

HRESULT STDMETHODCALLTYPE Document_MemAlloc(IDocument *this, DWORD dwSize, BOOL bExeMemory, int *nPointer)
{
  IsAkelPadReady();
  {
    if (!bExeMemory)
    {
      if (*nPointer=(int)GlobalAlloc(GPTR, dwSize))
        return NOERROR;
    }
    else
    {
      if (*nPointer=(int)SendMessage(hMainWnd, AKD_GLOBALALLOC, GPTR, dwSize))
        return NOERROR;
    }
  }
  return E_OUTOFMEMORY;
}

HRESULT STDMETHODCALLTYPE Document_MemCopy(IDocument *this, int nPointer, VARIANT vpData, DWORD dwType, BOOL bExeMemory, int *nBytes)
{
  HRESULT hr=NOERROR;

  IsAkelPadReady();
  {
    DWORD dwNumber=0;
    unsigned char *lpString=NULL;
    int nStringLen;

    *nBytes=0;

    if (vpData.vt == VT_BOOL)
      dwNumber=vpData.boolVal;
    else if (vpData.vt == VT_I2)
      dwNumber=vpData.iVal;
    else if (vpData.vt == VT_I4)
      dwNumber=vpData.lVal;
    else if (vpData.vt == VT_R8)
      dwNumber=(DWORD)vpData.dblVal;
    else if (vpData.vt == VT_DISPATCH)
      dwNumber=(DWORD)vpData.pdispVal;
    else if (vpData.vt == VT_BSTR)
    {
      lpString=(unsigned char *)vpData.bstrVal;
      nStringLen=SysStringLen(vpData.bstrVal) + 1;
    }

    if (dwType == DT_ANSI)
    {
      if (!bExeMemory)
      {
        if (lpString)
        {
          if (nPointer)
            *nBytes=WideCharToMultiByte(CP_ACP, 0, (wchar_t *)lpString, nStringLen, (char *)nPointer, 0x7FFFFFFF, NULL, NULL);
          else
            *nBytes=WideCharToMultiByte(CP_ACP, 0, (wchar_t *)lpString, nStringLen, NULL, 0, NULL, NULL);
        }
      }
      else
      {
        char *szString;
        int nStringLenAnsi;

        if (lpString)
        {
          if (nPointer)
          {
            nStringLenAnsi=WideCharToMultiByte(CP_ACP, 0, (wchar_t *)lpString, nStringLen, NULL, 0, NULL, NULL);

            if (szString=(char *)GlobalAlloc(GPTR, nStringLenAnsi))
            {
              WideCharToMultiByte(CP_ACP, 0, (wchar_t *)lpString, nStringLen, szString, nStringLenAnsi, NULL, NULL);
              WriteProcessMemory(hAkelProcess, (void *)nPointer, szString, nStringLenAnsi, nBytes);
              GlobalFree((HGLOBAL)szString);
            }
            else hr=E_OUTOFMEMORY;
          }
          else *nBytes=WideCharToMultiByte(CP_ACP, 0, (wchar_t *)lpString, nStringLen, NULL, 0, NULL, NULL);
        }
      }
    }
    else if (dwType == DT_UNICODE)
    {
      if (!bExeMemory)
      {
        if (lpString)
        {
          if (nPointer)
            memcpy((void *)nPointer, lpString, nStringLen * sizeof(wchar_t));
          *nBytes=nStringLen * sizeof(wchar_t);
        }
      }
      else
      {
        if (lpString)
        {
          if (nPointer)
          {
            WriteProcessMemory(hAkelProcess, (void *)nPointer, lpString, nStringLen * sizeof(wchar_t), nBytes);
          }
          else *nBytes=nStringLen * sizeof(wchar_t);
        }
      }
    }
    else if (dwType == DT_WORD)
    {
      if (!bExeMemory)
      {
        if (nPointer)
          memcpy((void *)nPointer, &dwNumber, sizeof(WORD));
        *nBytes=sizeof(DWORD);
      }
      else
      {
        if (nPointer)
        {
          WriteProcessMemory(hAkelProcess, (void *)nPointer, &dwNumber, sizeof(WORD), nBytes);
        }
        else *nBytes=sizeof(DWORD);
      }
    }
    else if (dwType == DT_DWORD)
    {
      if (!bExeMemory)
      {
        if (nPointer)
          memcpy((void *)nPointer, &dwNumber, sizeof(DWORD));
        *nBytes=sizeof(DWORD);
      }
      else
      {
        if (nPointer)
        {
          WriteProcessMemory(hAkelProcess, (void *)nPointer, &dwNumber, sizeof(DWORD), nBytes);
        }
        else *nBytes=sizeof(DWORD);
      }
    }
  }
  return hr;
}

HRESULT STDMETHODCALLTYPE Document_MemRead(IDocument *this, int nPointer, DWORD dwType, BOOL bExeMemory, VARIANT *vpData)
{
  HRESULT hr=NOERROR;

  IsAkelPadReady();
  {
    wchar_t *wszString;
    int nStringLen;

    VariantInit(vpData);

    if (dwType == DT_ANSI)
    {
      if (!bExeMemory)
      {
        nStringLen=lstrlenA((char *)nPointer);

        if (wszString=(wchar_t *)GlobalAlloc(GPTR, nStringLen * sizeof(wchar_t) + 2))
        {
          MultiByteToWideChar(CP_ACP, 0, (char *)nPointer, nStringLen + 1, wszString, nStringLen + 1);
          vpData->vt=VT_BSTR;
          if (!(vpData->bstrVal=SysAllocString(wszString)))
            hr=E_OUTOFMEMORY;
          GlobalFree((HGLOBAL)wszString);
        }
        else hr=E_OUTOFMEMORY;
      }
      else
      {
        char *szString;
        wchar_t *wszString;
        int nStringLen;

        if (nStringLen=SendMessage(hMainWnd, AKD_STRLENA, (WPARAM)nPointer, 0))
        {
          if (szString=(char *)GlobalAlloc(GPTR, nStringLen + 1))
          {
            if (ReadProcessMemory(hAkelProcess, (void *)nPointer, szString, nStringLen + 1, NULL))
            {
              if (wszString=(wchar_t *)GlobalAlloc(GPTR, nStringLen * sizeof(wchar_t) + 2))
              {
                MultiByteToWideChar(CP_ACP, 0, szString, nStringLen + 1, wszString, nStringLen + 1);
                vpData->vt=VT_BSTR;
                if (!(vpData->bstrVal=SysAllocString(wszString)))
                  hr=E_OUTOFMEMORY;
                GlobalFree((HGLOBAL)wszString);
              }
              else hr=E_OUTOFMEMORY;
            }
            GlobalFree((HGLOBAL)wszString);
          }
          else hr=E_OUTOFMEMORY;
        }
      }
    }
    else if (dwType == DT_UNICODE)
    {
      if (!bExeMemory)
      {
        vpData->vt=VT_BSTR;
        if (!(vpData->bstrVal=SysAllocString((wchar_t *)nPointer)))
          hr=E_OUTOFMEMORY;
      }
      else
      {
        wchar_t *wszString;
        int nStringLen;

        if (nStringLen=SendMessage(hMainWnd, AKD_STRLENW, (WPARAM)nPointer, 0))
        {
          if (wszString=(wchar_t *)GlobalAlloc(GPTR, nStringLen * sizeof(wchar_t) + 2))
          {
            if (ReadProcessMemory(hAkelProcess, (void *)nPointer, wszString, nStringLen * sizeof(wchar_t) + 2, NULL))
            {
              vpData->vt=VT_BSTR;
              if (!(vpData->bstrVal=SysAllocString(wszString)))
                hr=E_OUTOFMEMORY;
            }
            GlobalFree((HGLOBAL)wszString);
          }
          else hr=E_OUTOFMEMORY;
        }
      }
    }
    else if (dwType == DT_WORD)
    {
      if (!bExeMemory)
      {
        vpData->vt=VT_I2;
        vpData->iVal=LOWORD(*(DWORD *)nPointer);
      }
      else
      {
        vpData->vt=VT_I2;
        ReadProcessMemory(hAkelProcess, (void *)nPointer, &vpData->vt, sizeof(WORD), NULL);
      }
    }
    else if (dwType == DT_DWORD)
    {
      if (!bExeMemory)
      {
        vpData->vt=VT_I4;
        vpData->lVal=*(DWORD *)nPointer;
      }
      else
      {
        vpData->vt=VT_I4;
        ReadProcessMemory(hAkelProcess, (void *)nPointer, &vpData->lVal, sizeof(DWORD), NULL);
      }
    }
  }
  return hr;
}

HRESULT STDMETHODCALLTYPE Document_MemFree(IDocument *this, BOOL bExeMemory, int nPointer)
{
  IsAkelPadReady();
  {
    if (!bExeMemory)
    {
      if (nPointer) GlobalFree((HGLOBAL)nPointer);
    }
    else
    {
      if (nPointer) SendMessage(hMainWnd, AKD_GLOBALFREE, (WPARAM)nPointer, 0);
    }
  }
  return NOERROR;
}

HRESULT STDMETHODCALLTYPE Document_WindowRegisterClass(IDocument *this, BSTR wpClassName, WORD *wAtom)
{
  if (bOldWindows)
  {
    WNDCLASSA wndclassA={0};
    char szClassName[MAX_PATH];

    WideCharToMultiByte(CP_ACP, 0, wpClassName, -1, szClassName, MAX_PATH, NULL, NULL);

    wndclassA.style        =CS_HREDRAW|CS_VREDRAW;
    wndclassA.lpfnWndProc  =CallbackProcA;
    wndclassA.cbClsExtra   =0;
    wndclassA.cbWndExtra   =DLGWINDOWEXTRA;
    wndclassA.hInstance    =hInstanceDLL;
    wndclassA.hIcon        =NULL;
    wndclassA.hCursor      =LoadCursor(NULL, IDC_ARROW);
    wndclassA.hbrBackground=(HBRUSH)(COLOR_BTNFACE + 1);
    wndclassA.lpszMenuName =NULL;
    wndclassA.lpszClassName=szClassName;
    *wAtom=RegisterClassA(&wndclassA);
  }
  else
  {
    WNDCLASSW wndclass={0};

    wndclass.style        =CS_HREDRAW|CS_VREDRAW;
    wndclass.lpfnWndProc  =CallbackProcW;
    wndclass.cbClsExtra   =0;
    wndclass.cbWndExtra   =DLGWINDOWEXTRA;
    wndclass.hInstance    =hInstanceDLL;
    wndclass.hIcon        =NULL;
    wndclass.hCursor      =LoadCursor(NULL, IDC_ARROW);
    wndclass.hbrBackground=(HBRUSH)(COLOR_BTNFACE + 1);
    wndclass.lpszMenuName =NULL;
    wndclass.lpszClassName=wpClassName;
    *wAtom=RegisterClassW(&wndclass);
  }
  return NOERROR;
}

HRESULT STDMETHODCALLTYPE Document_WindowUnregisterClass(IDocument *this, BSTR wpClassName, BOOL *bResult)
{
  if (bOldWindows)
  {
    char szClassName[MAX_PATH];

    WideCharToMultiByte(CP_ACP, 0, wpClassName, -1, szClassName, MAX_PATH, NULL, NULL);

    *bResult=UnregisterClassA(szClassName, hInstanceDLL);
  }
  else
  {
    *bResult=UnregisterClassW(wpClassName, hInstanceDLL);
  }
  return NOERROR;
}

HRESULT STDMETHODCALLTYPE Document_WindowGetMessage(IDocument *this)
{
  MSG msg;
  BOOL bMsgStatus;

  if (bOldWindows)
  {
    while ((bMsgStatus=GetMessageA(&msg, NULL, 0, 0)) && bMsgStatus != -1)
    {
      if (g_hWndDialog && IsDialogMessageA(g_hWndDialog, &msg))
      {
        if (msg.message >= WM_KEYFIRST && msg.message <= WM_KEYLAST)
        {
          SendMessageA(g_hWndDialog, msg.message, msg.wParam, msg.lParam);
        }
      }
      else
      {
        TranslateMessage(&msg);
        DispatchMessageA(&msg);
      }
    }
  }
  else
  {
    while ((bMsgStatus=GetMessageW(&msg, NULL, 0, 0)) && bMsgStatus != -1)
    {
      if (g_hWndDialog && IsDialogMessageW(g_hWndDialog, &msg))
      {
        if (msg.message >= WM_KEYFIRST && msg.message <= WM_KEYLAST)
        {
          SendMessageW(g_hWndDialog, msg.message, msg.wParam, msg.lParam);
        }
      }
      else
      {
        TranslateMessage(&msg);
        DispatchMessageW(&msg);
      }
    }
  }
  return NOERROR;
}

HRESULT CallScriptFunction(IDispatch *objFunction, HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, LRESULT *lResult)
{
  OLECHAR *wpFuncName;
  DISPID dispid;
  DISPPARAMS dispp;
  VARIANT vtArg[6];
  VARIANT vtResult;
  HRESULT hr=NOERROR;

  wpFuncName=(OLECHAR *)L"call";

  if (!(hr=objFunction->lpVtbl->GetIDsOfNames(objFunction, &IID_NULL, &wpFuncName, 1, LOCALE_USER_DEFAULT, &dispid)))
  {
    VariantInit(&vtArg[0]);
    vtArg[0].lVal=lParam;
    vtArg[0].vt=VT_I4;
    VariantInit(&vtArg[1]);
    vtArg[1].lVal=wParam;
    vtArg[1].vt=VT_I4;
    VariantInit(&vtArg[2]);
    vtArg[2].lVal=uMsg;
    vtArg[2].vt=VT_I4;
    VariantInit(&vtArg[3]);
    vtArg[3].lVal=(DWORD)hWnd;
    vtArg[3].vt=VT_I4;
    VariantInit(&vtArg[4]);

    memset(&dispp, 0, sizeof(DISPPARAMS));
    dispp.cArgs=5;
    dispp.rgvarg=vtArg;

    if (!(hr=objFunction->lpVtbl->Invoke(objFunction, dispid, &IID_NULL, LOCALE_USER_DEFAULT, DISPATCH_METHOD, &dispp, &vtResult, 0, 0)))
    {
      if (lResult) *lResult=vtResult.lVal;
    }
  }
  return hr;
}

LRESULT CALLBACK CallbackProcA(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  LRESULT lResult;

  if (uMsg == WM_CREATE)
  {
    if (lParam)
    {
      CREATESTRUCT *cs=(CREATESTRUCT *)lParam;
      IDispatch *objCallback=(IDispatch *)cs->lpCreateParams;

      g_objCallback=objCallback;
      g_hWndDialog=hWnd;
    }
    SendMessage(hWnd, WM_SETICON, (WPARAM)ICON_BIG, (LPARAM)hIconDLL);
  }

  if (g_objCallback)
  {
    Sleep(0);

    if (!(CallScriptFunction(g_objCallback, hWnd, uMsg, wParam, lParam, &lResult)))
    {
      if (lResult) return lResult;
    }
  }

  if (uMsg == WM_DESTROY)
  {
    g_objCallback=NULL;
    g_hWndDialog=NULL;
  }
  return DefWindowProcA(hWnd, uMsg, wParam, lParam);
}

LRESULT CALLBACK CallbackProcW(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  LRESULT lResult;

  if (uMsg == WM_CREATE)
  {
    if (lParam)
    {
      CREATESTRUCT *cs=(CREATESTRUCT *)lParam;
      IDispatch *objCallback=(IDispatch *)cs->lpCreateParams;

      g_objCallback=objCallback;
      g_hWndDialog=hWnd;
    }
    SendMessage(hWnd, WM_SETICON, (WPARAM)ICON_BIG, (LPARAM)hIconDLL);
  }

  if (g_objCallback)
  {
    Sleep(0);

    if (!(CallScriptFunction(g_objCallback, hWnd, uMsg, wParam, lParam, &lResult)))
    {
      if (lResult) return lResult;
    }
  }

  if (uMsg == WM_DESTROY)
  {
    g_objCallback=NULL;
    g_hWndDialog=NULL;
  }
  return DefWindowProcW(hWnd, uMsg, wParam, lParam);
}
