#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <objbase.h>
#include <commctrl.h>
#include <richedit.h>
#include <initguid.h>
#include "AkelEdit.h"
#include "AkelDLL.h"
#include "StackFunc.h"
#include "ConvFunc.h"
#include "ISystemFunction.h"
#include "IDocument.h"
#include "IServer.h"
#include "Scripts.h"
#include "Resources\Resource.h"


//GUID

// IDocument object's GUID
// {DB045777-BAFF-416b-AA8E-A154E6A64A88}
DEFINE_GUID(CLSID_IDocument, 0xdb045777, 0xbaff, 0x416b, 0xaa, 0x8e, 0xa1, 0x54, 0xe6, 0xa6, 0x4a, 0x88);

// IDocument type library's GUID
// {DB4185E2-3423-409f-B875-4ED126BF32C2}
DEFINE_GUID(CLSID_TypeLib, 0xdb4185e2, 0x3423, 0x409f, 0xb8, 0x75, 0x4e, 0xd1, 0x26, 0xbf, 0x32, 0xc2);

// IDocument VTable's GUID
// {DBE0F625-75CE-44f7-B1EC-BCC83D5177BE}
DEFINE_GUID(IID_IDocument, 0xdbe0f625, 0x75ce, 0x44f7, 0xb1, 0xec, 0xbc, 0xc8, 0x3d, 0x51, 0x77, 0xbe);

// ISystemFunction VTable's GUID
// {DBB12030-4BEF-4bbe-AAD1-0A01B27084F9}
DEFINE_GUID(IID_ISystemFunction, 0xdbb12030, 0x4bef, 0x4bbe, 0xaa, 0xd1, 0xa, 0x1, 0xb2, 0x70, 0x84, 0xf9);

//Global variables
DWORD g_nObjs;
DWORD g_nLocks;
IClassFactory MyIClassFactoryObj;

const IClassFactoryVtbl MyIClassFactoryVtbl={
  classQueryInterface,
  classAddRef,
  classRelease,
  classCreateInstance,
  classLockServer
};


//// IClassFactory

ULONG STDMETHODCALLTYPE classAddRef(IClassFactory *this)
{
  InterlockedIncrement(&g_nObjs);
  return 1;
}

HRESULT STDMETHODCALLTYPE classQueryInterface(IClassFactory *this, REFIID factoryGuid, void **ppv)
{
  if (IsEqualIID(factoryGuid, &IID_IUnknown) || IsEqualIID(factoryGuid, &IID_IClassFactory))
  {
    this->lpVtbl->AddRef(this);
    *ppv=this;
    return NOERROR;
  }
  *ppv=NULL;
  return E_NOINTERFACE;
}

ULONG STDMETHODCALLTYPE classRelease(IClassFactory *this)
{
  return InterlockedDecrement(&g_nObjs);
}

HRESULT STDMETHODCALLTYPE classCreateInstance(IClassFactory *this, IUnknown *punkOuter, REFIID vTableGuid, void **objHandle)
{
  IDocument *thisobj;
  HRESULT hr;

  *objHandle=NULL;
  if (punkOuter) return CLASS_E_NOAGGREGATION;

  if (!(thisobj=(IDocument *)GlobalAlloc(GMEM_FIXED, sizeof(IRealDocument))))
    return E_OUTOFMEMORY;

  thisobj->lpVtbl=(IDocumentVtbl *)&MyIDocumentVtbl;
  ((IRealDocument *)thisobj)->dwCount=1;

  hr=MyIDocumentVtbl.Document_QueryInterface(thisobj, vTableGuid, objHandle);
  MyIDocumentVtbl.Document_Release(thisobj);
  if (hr == S_OK) InterlockedIncrement(&g_nObjs);

  return hr;
}

HRESULT STDMETHODCALLTYPE classLockServer(IClassFactory *this, BOOL flock)
{
  if (flock)
    InterlockedIncrement(&g_nLocks);
  else
    InterlockedDecrement(&g_nLocks);
  return NOERROR;
}


//// Extern functions

HRESULT WINAPI DllGetClassObject(REFCLSID objGuid, REFIID factoryGuid, void **factoryHandle)
{
  if (IsEqualCLSID(objGuid, &CLSID_IDocument))
  {
    return classQueryInterface(&MyIClassFactoryObj, factoryGuid, factoryHandle);
  }
  *factoryHandle=NULL;
  return CLASS_E_CLASSNOTAVAILABLE;
}

HRESULT WINAPI DllCanUnloadNow(void)
{
 if (g_nObjs || g_nLocks)
   return S_FALSE;
 else
   return S_OK;
}

HRESULT WINAPI DllRegisterServer()
{
  const char szNameTypeLib[]="Scripts.dll";
  const char szNameProgID[]="AkelPad.document";
  const char szObjectDescription[]="AkelPad COM component";

  return RegisterServerA(hInstanceDLL, (REFCLSID)&CLSID_IDocument, szNameTypeLib, szNameProgID, szObjectDescription);
}

HRESULT WINAPI DllUnregisterServer()
{
  const char szNameProgID[]="AkelPad.document";

  return UnregisterServerA((REFCLSID)&CLSID_IDocument, (REFGUID)&CLSID_TypeLib, szNameProgID);
}

void stringFromCLSIDA(char *szString, REFCLSID clsid)
{
  wsprintfA(szString, "{%08lX-%04X-%04X-%02X%02X-%02X%02X%02X%02X%02X%02X}",
    clsid->Data1, clsid->Data2, clsid->Data3,
    clsid->Data4[0], clsid->Data4[1], clsid->Data4[2], clsid->Data4[3],
    clsid->Data4[4], clsid->Data4[5], clsid->Data4[6], clsid->Data4[7]);
}

HRESULT IsServerRegisteredA(HINSTANCE hInstance, REFCLSID clsid, const char *pNameTypeLib, const char *pNameProgID, const char *pObjectDescription, char *szDllFile, int nDllFileLen)
{
  char szGUID[64];
  char szDLL[MAX_PATH];
  char *pDllName;
  HKEY hKey;
  DWORD dwSize;
  BSTR bstrDll;
  HRESULT hr=S_OK;

  if (szDllFile) szDllFile[0]='\0';

  //{DB045777-BAFF-416b-AA8E-A154E6A64A88}
  stringFromCLSIDA(szGUID, clsid);

  //Get file name Scripts.dll
  GetModuleFileNameA(hInstance, szDLL, MAX_PATH);
  pDllName=GetFileNameA(szDLL);

  //HKEY_LOCAL_MACHINE\SOFTWARE\Classes\AkelPad.document
  wsprintfA(szBuffer, "Software\\Classes\\%s", pNameProgID);
  if (RegOpenKeyExA(HKEY_LOCAL_MACHINE, szBuffer, 0, KEY_READ, &hKey) == ERROR_SUCCESS)
  {
    dwSize=BUFFER_SIZE;
    RegQueryValueExA(hKey, NULL, NULL, NULL, szBuffer, &dwSize);
    RegCloseKey(hKey);
  }
  else return S_FALSE;

  if (lstrcmpiA(pObjectDescription, szBuffer))
    return S_FALSE;

  //HKEY_LOCAL_MACHINE\SOFTWARE\Classes\AkelPad.document\CLSID
  wsprintfA(szBuffer, "Software\\Classes\\%s\\CLSID", pNameProgID);
  if (RegOpenKeyExA(HKEY_LOCAL_MACHINE, szBuffer, 0, KEY_READ, &hKey) == ERROR_SUCCESS)
  {
    dwSize=BUFFER_SIZE;
    RegQueryValueExA(hKey, NULL, NULL, NULL, szBuffer, &dwSize);
    RegCloseKey(hKey);
  }
  else return S_FALSE;

  if (lstrcmpiA(szGUID, szBuffer))
    return S_FALSE;

  //HKEY_LOCAL_MACHINE\SOFTWARE\Classes\CLSID\{DB045777-BAFF-416B-AA8E-A154E6A64A88}
  wsprintfA(szBuffer, "Software\\Classes\\CLSID\\%s", szGUID);
  if (RegOpenKeyExA(HKEY_LOCAL_MACHINE, szBuffer, 0, KEY_READ, &hKey) == ERROR_SUCCESS)
  {
    dwSize=BUFFER_SIZE;
    RegQueryValueExA(hKey, NULL, NULL, NULL, szBuffer, &dwSize);
    RegCloseKey(hKey);
  }
  else return S_FALSE;

  if (lstrcmpiA(pObjectDescription, szBuffer))
    return S_FALSE;

  //HKEY_LOCAL_MACHINE\SOFTWARE\Classes\CLSID\{DB045777-BAFF-416B-AA8E-A154E6A64A88}\InprocServer32
  wsprintfA(szBuffer, "Software\\Classes\\CLSID\\%s\\InprocServer32", szGUID);
  if (RegOpenKeyExA(HKEY_LOCAL_MACHINE, szBuffer, 0, KEY_READ, &hKey) == ERROR_SUCCESS)
  {
    dwSize=BUFFER_SIZE;
    RegQueryValueExA(hKey, NULL, NULL, NULL, szBuffer, &dwSize);
    RegCloseKey(hKey);
  }
  else return S_FALSE;

  if (lstrcmpiA(pDllName, GetFileNameA(szBuffer)) || !FileExistsA(szBuffer))
    return S_FALSE;

  //HKEY_LOCAL_MACHINE\SOFTWARE\Classes\CLSID\{DB045777-BAFF-416B-AA8E-A154E6A64A88}\ProgID
  wsprintfA(szBuffer, "Software\\Classes\\CLSID\\%s\\ProgID", szGUID);
  if (RegOpenKeyExA(HKEY_LOCAL_MACHINE, szBuffer, 0, KEY_READ, &hKey) == ERROR_SUCCESS)
  {
    dwSize=BUFFER_SIZE;
    RegQueryValueExA(hKey, NULL, NULL, NULL, szBuffer, &dwSize);
    RegCloseKey(hKey);
  }
  else return S_FALSE;

  if (lstrcmpiA(pNameProgID, szBuffer))
    return S_FALSE;

  if (bstrDll=SysAllocStringLen(NULL, MAX_PATH))
  {
    hr=QueryPathOfRegTypeLib((REFGUID)&CLSID_TypeLib, 1, 0, LOCALE_NEUTRAL, &bstrDll);

    if (hr == S_OK)
    {
      WideCharToMultiByte(CP_ACP, 0, bstrDll, -1, szBuffer, MAX_PATH, NULL, NULL);
      if (lstrcmpiA(pDllName, GetFileNameA(szBuffer)) || !FileExistsA(szBuffer))
        hr=S_FALSE;
      if (szDllFile) lstrcpynA(szDllFile, szBuffer, nDllFileLen);
    }
    SysFreeString(bstrDll);
  }
  else return S_FALSE;

  return hr;
}

HRESULT RegisterServerA(HINSTANCE hInstance, REFCLSID clsid, const char *pNameTypeLib, const char *pNameProgID, const char *pObjectDescription)
{
  char szGUID[64];
  char szDLL[MAX_PATH];
  HKEY hRootKey;
  HKEY hKey;
  HKEY hKey2;
  HKEY hKey3;
  HRESULT hr=S_OK;

  GetModuleFileNameA(hInstance, szDLL, MAX_PATH);

  if (RegOpenKeyExA(HKEY_LOCAL_MACHINE, "Software\\Classes", 0, KEY_WRITE, &hRootKey) == ERROR_SUCCESS)
  {
    if (RegCreateKeyExA(hRootKey, pNameProgID, 0, 0, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, 0, &hKey, NULL) == ERROR_SUCCESS)
    {
      RegSetValueEx(hKey, NULL, 0, REG_SZ, (const BYTE *)pObjectDescription, lstrlenA(pObjectDescription) + 1);

      if (RegCreateKeyExA(hKey, "CLSID", 0, 0, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, 0, &hKey2, NULL) == ERROR_SUCCESS)
      {
        stringFromCLSIDA(szGUID, clsid);
        RegSetValueEx(hKey2, NULL, 0, REG_SZ, (const BYTE *)szGUID, lstrlenA(szGUID) + 1);
        RegCloseKey(hKey2);
      }
      else hr=S_FALSE;

      RegCloseKey(hKey);
    }
    else hr=S_FALSE;

    if (hr == S_OK && RegOpenKeyExA(hRootKey, "CLSID", 0, KEY_ALL_ACCESS, &hKey) == ERROR_SUCCESS)
    {
      if (RegCreateKeyExA(hKey, szGUID, 0, 0, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, 0, &hKey2, NULL) == ERROR_SUCCESS)
      {
        RegSetValueEx(hKey2, NULL, 0, REG_SZ, (const BYTE *)pObjectDescription, lstrlenA(pObjectDescription) + 1);

        if (RegCreateKeyExA(hKey2, "InprocServer32", 0, 0, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, 0, &hKey3, NULL) == ERROR_SUCCESS)
        {
          RegSetValueEx(hKey3, NULL, 0, REG_SZ, (const BYTE *)szDLL, lstrlenA(szDLL) + 1);
          RegCloseKey(hKey3);
        }
        else hr=S_FALSE;

        if (hr == S_OK && RegCreateKeyExA(hKey2, "ProgID", 0, 0, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, 0, &hKey3, NULL) == ERROR_SUCCESS)
        {
          RegSetValueEx(hKey3, NULL, 0, REG_SZ, (const BYTE *)pNameProgID, lstrlenA(pNameProgID) + 1);
          RegCloseKey(hKey3);
        }
        else hr=S_FALSE;

        RegCloseKey(hKey2);
      }
      else hr=S_FALSE;

      RegCloseKey(hKey);
    }
    else hr=S_FALSE;

    RegCloseKey(hRootKey);
  }
  else hr=S_FALSE;

  if (hr == S_OK)
  {
    wchar_t wszGUID[MAX_PATH];
    ITypeLib *pTypeLib;
    int i;

    for (i=lstrlenA(szDLL) - 1; i >= 0; --i)
      if (szDLL[i] == '\\') break;
    lstrcpyA(szDLL + i + 1, pNameTypeLib);

    MultiByteToWideChar(CP_ACP, 0, szDLL, -1, wszGUID, MAX_PATH);
    if ((hr=LoadTypeLib(wszGUID, &pTypeLib)) == S_OK)
    {
      hr=RegisterTypeLib(pTypeLib, wszGUID, NULL);
      pTypeLib->lpVtbl->Release(pTypeLib);
    }
  }
  return hr;
}

HRESULT UnregisterServerA(REFCLSID clsid, REFGUID typelib, const char *pNameProgID)
{
  char szGUID[64];
  HKEY hRootKey;
  HKEY hKey;
  HKEY hKey2;

  stringFromCLSIDA(szGUID, clsid);

  if (RegOpenKeyExA(HKEY_LOCAL_MACHINE, "Software\\Classes", 0, KEY_WRITE, &hRootKey) == ERROR_SUCCESS)
  {
    if (RegOpenKeyExA(hRootKey, pNameProgID, 0, KEY_ALL_ACCESS, &hKey) == ERROR_SUCCESS)
    {
      RegDeleteKeyA(hKey, "CLSID");
      RegCloseKey(hKey);
      RegDeleteKeyA(hRootKey, pNameProgID);
    }
    if (RegOpenKeyExA(hRootKey, "CLSID", 0, KEY_ALL_ACCESS, &hKey) == ERROR_SUCCESS)
    {
      if (RegOpenKeyExA(hKey, szGUID, 0, KEY_ALL_ACCESS, &hKey2) == ERROR_SUCCESS)
      {
        RegDeleteKeyA(hKey2, "InprocServer32");
        RegDeleteKeyA(hKey2, "ProgID");
        RegCloseKey(hKey2);
        RegDeleteKeyA(hKey, szGUID);
      }
      RegCloseKey(hKey);
    }
    RegCloseKey(hRootKey);

    UnRegisterTypeLib(typelib, 1, 0, LOCALE_NEUTRAL, SYS_WIN32);
  }
  return S_OK;
}
