#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <objbase.h>
#include <commctrl.h>
#include <richedit.h>
#include "AkelEdit.h"
#include "AkelDLL.h"
#include "StackFunc.h"
#include "ConvFunc.h"
#include "ISystemFunction.h"
#include "IDocument.h"
#include "IServer.h"
#include "Scripts.h"
#include "Resources\Resource.h"

//Include stack functions
#define StackInsertAfter
#define StackInsertBefore
#define StackInsertIndex
#define StackClear
#include "StackFunc.h"

//Global variables
char szPluginName[MAX_PATH];
wchar_t wszPluginName[MAX_PATH];
PLUGINDATA pdMainInfo;
HWND hMainWnd=NULL;
BOOL bOldWindows;
LANGID wLangSystem=0;
BOOL bInitCommon=FALSE;
BOOL bInitMain=FALSE;
HINSTANCE hInstanceDLL=NULL;
HINSTANCE hAkelInstance=NULL;
HANDLE hAkelProcess=NULL;
DWORD dwAkelProcessId=0;
DWORD dwArchitectVersion=0;
HWND hWndMainDlg=NULL;
RECT rcMainInitDialog={0};
RECT rcMainCurrentDialog={0};
BOOL bMainDialogRectSave=FALSE;
int nColumnWidth1=163;
int nColumnWidth2=109;
HICON hIconDLL=NULL;
char szMutexInitName[]="AkelPad::Scripts::MutexInit";
wchar_t wszMutexInitName[]=L"AkelPad::Scripts::MutexInit";
char szMutexExecName[]="AkelPad::Scripts::MutexExec";
wchar_t wszMutexExecName[]=L"AkelPad::Scripts::MutexExec";
char szScriptsDir[MAX_PATH];
wchar_t wszScriptsDir[MAX_PATH];
char szBuffer[BUFFER_SIZE];
wchar_t wszBuffer[BUFFER_SIZE];
char szCmdLine[32768];
wchar_t wszCmdLine[32768];
HSTACK hScriptsStack={0};
HANDLE hExecThread=NULL;
DWORD dwExecThreadId=0;
int nRegisterMode=RM_MANUAL;
char szServerFile[MAX_PATH];
BOOL bServerRegistered=FALSE;

//Identification
void __declspec(dllexport) DllAkelPadID(PLUGINVERSION *pv)
{
  pv->dwAkelDllVersion=AKELDLL;
  pv->dwExeMinVersion3x=MAKE_IDENTIFIER(3, 7, 4, 0);
  pv->dwExeMinVersion4x=MAKE_IDENTIFIER(4, 4, 0, 0);
  pv->pPluginName="Scripts";
}

//Plugin extern function
void __declspec(dllexport) Main(PLUGINDATA *pd)
{
  if (pd->lpbAutoLoad)
  {
    *pd->lpbAutoLoad=TRUE;
    return;
  }

  if (!bInitCommon) InitCommon(pd);
  if (!bInitMain) InitMain();

  if (pd->lParam)
  {
    if (IsExtCallParamValid(pd->lParam, 1))
    {
      if (bOldWindows)
      {
        if (GetExtCallParam(pd->lParam, 1) == DLLA_SCRIPTS_EXEC)
        {
          char *pScript=NULL;
          char *pArguments=NULL;

          if (IsExtCallParamValid(pd->lParam, 2))
            pScript=(char *)GetExtCallParam(pd->lParam, 2);
          if (IsExtCallParamValid(pd->lParam, 3))
            pArguments=(char *)GetExtCallParam(pd->lParam, 3);

          if (pScript)
          {
            ExecScriptA(pScript, pArguments);
          }
        }
        else if (GetExtCallParam(pd->lParam, 1) == DLLA_SCRIPTS_EDIT)
        {
          char *pScript=NULL;

          if (IsExtCallParamValid(pd->lParam, 2))
            pScript=(char *)GetExtCallParam(pd->lParam, 2);

          if (pScript)
          {
            EditScriptA(pScript);
          }
        }
      }
      else
      {
        if (GetExtCallParam(pd->lParam, 1) == DLLA_SCRIPTS_EXEC)
        {
          wchar_t *wpScript=NULL;
          wchar_t *wpArguments=NULL;

          if (IsExtCallParamValid(pd->lParam, 2))
            wpScript=(wchar_t *)GetExtCallParam(pd->lParam, 2);
          if (IsExtCallParamValid(pd->lParam, 3))
            wpArguments=(wchar_t *)GetExtCallParam(pd->lParam, 3);

          if (wpScript)
          {
            ExecScriptW(wpScript, wpArguments);
          }
        }
        else if (GetExtCallParam(pd->lParam, 1) == DLLA_SCRIPTS_EDIT)
        {
          wchar_t *wpScript=NULL;

          if (IsExtCallParamValid(pd->lParam, 2))
            wpScript=(wchar_t *)GetExtCallParam(pd->lParam, 2);

          if (wpScript)
          {
            EditScriptW(wpScript);
          }
        }
      }
    }
  }
  else
  {
    if (!pd->bOnStart)
    {
      if (bOldWindows)
        DialogBoxA(hInstanceDLL, MAKEINTRESOURCEA(IDD_SCRIPTS), hMainWnd, (DLGPROC)MainDlgProcA);
      else
        DialogBoxW(hInstanceDLL, MAKEINTRESOURCEW(IDD_SCRIPTS), hMainWnd, (DLGPROC)MainDlgProcW);
    }
  }

  //Stay in memory, and show as active
  pd->nUnload=UD_NONUNLOAD_ACTIVE;
}

BOOL CALLBACK MainDlgProcA(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static HWND hWndList;
  static HWND hWndChangeListGroup;
  static HWND hWndExecButton;
  static HWND hWndEditButton;
  static HWND hWndHotkey;
  static HWND hWndAssignButton;
  static HWND hWndSettingsButton;
  static HWND hWndOKButton;
  static int nSelItem=-1;
  static BOOL bListChanged=FALSE;
  static DIALOGRESIZE drs[]={{&hWndList,            DRS_SIZE|DRS_X|DRS_Y},
                             {&hWndChangeListGroup, DRS_MOVE|DRS_X},
                             {&hWndExecButton,      DRS_MOVE|DRS_X},
                             {&hWndEditButton,      DRS_MOVE|DRS_X},
                             {&hWndHotkey,          DRS_MOVE|DRS_X},
                             {&hWndAssignButton,    DRS_MOVE|DRS_X},
                             {&hWndSettingsButton,  DRS_MOVE|DRS_X},
                             {&hWndOKButton,        DRS_MOVE|DRS_X|DRS_Y},
                             {0, 0}};

  if (uMsg == WM_INITDIALOG)
  {
    LVCOLUMNA lvc;

    hWndMainDlg=hDlg;

    SendMessage(hDlg, WM_SETICON, (WPARAM)ICON_BIG, (LPARAM)hIconDLL);
    hWndList=GetDlgItem(hDlg, IDC_SCRIPTS_LIST);
    hWndChangeListGroup=GetDlgItem(hDlg, IDC_CHANGELIST_GROUP);
    hWndExecButton=GetDlgItem(hDlg, IDC_EXEC);
    hWndEditButton=GetDlgItem(hDlg, IDC_EDIT);
    hWndHotkey=GetDlgItem(hDlg, IDC_HOTKEY);
    hWndAssignButton=GetDlgItem(hDlg, IDC_ASSIGN);
    hWndSettingsButton=GetDlgItem(hDlg, IDC_SETTINGS);
    hWndOKButton=GetDlgItem(hDlg, IDOK);

    EnableWindow(hWndExecButton, FALSE);
    EnableWindow(hWndEditButton, FALSE);
    EnableWindow(hWndHotkey, FALSE);
    EnableWindow(hWndAssignButton, FALSE);
    SendMessage(hWndList, LVM_SETEXTENDEDLISTVIEWSTYLE, LVS_EX_FULLROWSELECT|LVS_EX_GRIDLINES, LVS_EX_FULLROWSELECT|LVS_EX_GRIDLINES);

    SetDlgItemTextA(hDlg, IDC_EXEC, GetLangStringA(wLangSystem, STRID_EXEC));
    SetDlgItemTextA(hDlg, IDC_EDIT, GetLangStringA(wLangSystem, STRID_EDIT));
    SetDlgItemTextA(hDlg, IDC_ASSIGN, GetLangStringA(wLangSystem, STRID_ASSIGN));
    SetDlgItemTextA(hDlg, IDC_SETTINGS, GetLangStringA(wLangSystem, STRID_SETTINGS));
    SetDlgItemTextA(hDlg, IDOK, GetLangStringA(wLangSystem, STRID_OK));

    SendMessage(hMainWnd, AKD_SETHOTKEYINPUT, (WPARAM)hWndHotkey, 0);

    //Columns
    lvc.mask=LVCF_TEXT|LVCF_WIDTH|LVCF_SUBITEM;
    lvc.pszText=GetLangStringA(wLangSystem, STRID_SCRIPT);
    lvc.cx=nColumnWidth1;
    lvc.iSubItem=LVI_SCRIPT_FILE;
    SendMessage(hWndList, LVM_INSERTCOLUMNA, LVI_SCRIPT_FILE, (LPARAM)&lvc);

    lvc.mask=LVCF_TEXT|LVCF_WIDTH|LVCF_SUBITEM;
    lvc.pszText=GetLangStringA(wLangSystem, STRID_HOTKEY);
    lvc.cx=nColumnWidth2;
    lvc.iSubItem=LVI_SCRIPT_HOTKEY;
    SendMessage(hWndList, LVM_INSERTCOLUMNA, LVI_SCRIPT_HOTKEY, (LPARAM)&lvc);

    FillScriptListA(hWndList);
  }
  else if (uMsg == WM_NOTIFY)
  {
    if (wParam == IDC_SCRIPTS_LIST)
    {
      if (((NMLISTVIEW *)lParam)->hdr.code == LVN_ITEMCHANGED)
      {
        if (((NMLISTVIEW *)lParam)->uNewState & LVIS_STATEIMAGEMASK)
        {
          BOOL bNewState;
          BOOL bOldState;

          bNewState=((((NMLISTVIEW *)lParam)->uNewState & LVIS_STATEIMAGEMASK) >> 12) - 1;
          bOldState=((((NMLISTVIEW *)lParam)->uOldState & LVIS_STATEIMAGEMASK) >> 12) - 1;

          if (bNewState >=0 && bOldState >=0 && bNewState != bOldState)
            bListChanged=TRUE;
        }
        if (((NMLISTVIEW *)lParam)->uNewState & LVIS_SELECTED)
        {
          SendMessage(hWndHotkey, HKM_SETHOTKEY, ((NMLISTVIEW *)lParam)->lParam, 0);
          EnableWindow(hWndExecButton, TRUE);
          EnableWindow(hWndEditButton, TRUE);
          EnableWindow(hWndHotkey, TRUE);
          EnableWindow(hWndAssignButton, TRUE);
          nSelItem=((NMLISTVIEW *)lParam)->iItem;
        }
        if (((NMLISTVIEW *)lParam)->uOldState & LVIS_SELECTED)
        {
          SendMessage(hWndHotkey, HKM_SETHOTKEY, 0, 0);
          EnableWindow(hWndExecButton, FALSE);
          EnableWindow(hWndEditButton, FALSE);
          EnableWindow(hWndHotkey, FALSE);
          EnableWindow(hWndAssignButton, FALSE);
          nSelItem=-1;
        }
      }
      else if ((int)((NMHDR *)lParam)->code == NM_DBLCLK) PostMessage(hDlg, WM_COMMAND, IDC_EXEC, 0);
    }
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDC_ASSIGN)
    {
      LVITEMA lvi;
      WORD wHotkey;
      char szFile[MAX_PATH];
      char szHotkey[MAX_PATH];

      //Get macro name
      lvi.mask=LVIF_TEXT;
      lvi.pszText=szFile;
      lvi.cchTextMax=MAX_PATH;
      lvi.iItem=nSelItem;
      lvi.iSubItem=LVI_SCRIPT_FILE;
      SendMessage(hWndList, LVM_GETITEMA, 0, (LPARAM)&lvi);

      //Get macro hotkey
      wHotkey=(WORD)SendMessage(hWndHotkey, HKM_GETHOTKEY, 0, 0);

      //Register hotkey
      if (RegisterHotkeyA(szFile, wHotkey))
      {
        GetHotkeyStringA(wHotkey, szHotkey);

        lvi.mask=LVIF_PARAM;
        lvi.iItem=nSelItem;
        lvi.iSubItem=LVI_SCRIPT_FILE;
        lvi.lParam=wHotkey;
        SendMessage(hWndList, LVM_SETITEMA, 0, (LPARAM)&lvi);

        lvi.mask=LVIF_TEXT;
        lvi.pszText=szHotkey;
        lvi.iItem=nSelItem;
        lvi.iSubItem=LVI_SCRIPT_HOTKEY;
        SendMessage(hWndList, LVM_SETITEMA, 0, (LPARAM)&lvi);

        SendMessage(hWndHotkey, HKM_SETHOTKEY, wHotkey, 0);
        bListChanged=TRUE;
      }
      else SetFocus(hWndHotkey);

      return TRUE;
    }
    else if (LOWORD(wParam) == IDC_SETTINGS)
    {
      DialogBoxA(hInstanceDLL, MAKEINTRESOURCEA(IDD_SETTINGS), hDlg, (DLGPROC)SettingsDlgProcA);
    }
    else if (LOWORD(wParam) == IDOK ||
             LOWORD(wParam) == IDC_EXEC ||
             LOWORD(wParam) == IDC_EDIT ||
             LOWORD(wParam) == IDCANCEL)
    {
      LVITEMA lvi;
      char szName[MAX_PATH];
      int nWidth;

      nWidth=SendMessage(hWndList, LVM_GETCOLUMNWIDTH, LVI_SCRIPT_FILE, 0);
      if (nWidth != nColumnWidth1)
      {
        nColumnWidth1=nWidth;
        bMainDialogRectSave=TRUE;
      }
      nWidth=SendMessage(hWndList, LVM_GETCOLUMNWIDTH, LVI_SCRIPT_HOTKEY, 0);
      if (nWidth != nColumnWidth2)
      {
        nColumnWidth2=nWidth;
        bMainDialogRectSave=TRUE;
      }

      if (bListChanged || LOWORD(wParam) == IDOK)
      {
        SaveScriptListA(hWndList);
        SaveOptionsA(OF_MAIN|OF_WINDOWRECT);
      }

      if (LOWORD(wParam) == IDC_EXEC ||
          LOWORD(wParam) == IDC_EDIT)
      {
        //Get macro name
        lvi.mask=LVIF_TEXT;
        lvi.pszText=szName;
        lvi.cchTextMax=MAX_PATH;
        lvi.iItem=nSelItem;
        lvi.iSubItem=LVI_SCRIPT_FILE;
        SendMessage(hWndList, LVM_GETITEMA, 0, (LPARAM)&lvi);
      }

      //End dialog
      EndDialog(hDlg, 0);

      //Execute
      if (LOWORD(wParam) == IDC_EXEC)
        ExecScriptA(szName, NULL);
      else if (LOWORD(wParam) == IDC_EDIT)
        EditScriptA(szName);
      return TRUE;
    }
  }
  else if (uMsg == WM_CLOSE)
  {
    PostMessage(hDlg, WM_COMMAND, IDCANCEL, 0);
    return TRUE;
  }

  if (DialogResizeMessages(&drs[0], &rcMainInitDialog, &rcMainCurrentDialog, hDlg, uMsg, wParam, lParam))
    bMainDialogRectSave=TRUE;

  return FALSE;
}

BOOL CALLBACK MainDlgProcW(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static HWND hWndList;
  static HWND hWndChangeListGroup;
  static HWND hWndExecButton;
  static HWND hWndEditButton;
  static HWND hWndHotkey;
  static HWND hWndAssignButton;
  static HWND hWndSettingsButton;
  static HWND hWndOKButton;
  static int nSelItem=-1;
  static BOOL bListChanged=FALSE;
  static DIALOGRESIZE drs[]={{&hWndList,            DRS_SIZE|DRS_X|DRS_Y},
                             {&hWndChangeListGroup, DRS_MOVE|DRS_X},
                             {&hWndExecButton,      DRS_MOVE|DRS_X},
                             {&hWndEditButton,      DRS_MOVE|DRS_X},
                             {&hWndHotkey,          DRS_MOVE|DRS_X},
                             {&hWndAssignButton,    DRS_MOVE|DRS_X},
                             {&hWndSettingsButton,  DRS_MOVE|DRS_X},
                             {&hWndOKButton,        DRS_MOVE|DRS_X|DRS_Y},
                             {0, 0}};

  if (uMsg == WM_INITDIALOG)
  {
    LVCOLUMNW lvc;

    hWndMainDlg=hDlg;

    SendMessage(hDlg, WM_SETICON, (WPARAM)ICON_BIG, (LPARAM)hIconDLL);
    hWndList=GetDlgItem(hDlg, IDC_SCRIPTS_LIST);
    hWndChangeListGroup=GetDlgItem(hDlg, IDC_CHANGELIST_GROUP);
    hWndExecButton=GetDlgItem(hDlg, IDC_EXEC);
    hWndEditButton=GetDlgItem(hDlg, IDC_EDIT);
    hWndHotkey=GetDlgItem(hDlg, IDC_HOTKEY);
    hWndAssignButton=GetDlgItem(hDlg, IDC_ASSIGN);
    hWndSettingsButton=GetDlgItem(hDlg, IDC_SETTINGS);
    hWndOKButton=GetDlgItem(hDlg, IDOK);

    EnableWindow(hWndExecButton, FALSE);
    EnableWindow(hWndEditButton, FALSE);
    EnableWindow(hWndHotkey, FALSE);
    EnableWindow(hWndAssignButton, FALSE);
    SendMessage(hWndList, LVM_SETEXTENDEDLISTVIEWSTYLE, LVS_EX_FULLROWSELECT|LVS_EX_GRIDLINES, LVS_EX_FULLROWSELECT|LVS_EX_GRIDLINES);

    SetDlgItemTextW(hDlg, IDC_EXEC, GetLangStringW(wLangSystem, STRID_EXEC));
    SetDlgItemTextW(hDlg, IDC_EDIT, GetLangStringW(wLangSystem, STRID_EDIT));
    SetDlgItemTextW(hDlg, IDC_ASSIGN, GetLangStringW(wLangSystem, STRID_ASSIGN));
    SetDlgItemTextW(hDlg, IDC_SETTINGS, GetLangStringW(wLangSystem, STRID_SETTINGS));
    SetDlgItemTextW(hDlg, IDOK, GetLangStringW(wLangSystem, STRID_OK));

    SendMessage(hMainWnd, AKD_SETHOTKEYINPUT, (WPARAM)hWndHotkey, 0);

    //Columns
    lvc.mask=LVCF_TEXT|LVCF_WIDTH|LVCF_SUBITEM;
    lvc.pszText=GetLangStringW(wLangSystem, STRID_SCRIPT);
    lvc.cx=nColumnWidth1;
    lvc.iSubItem=LVI_SCRIPT_FILE;
    SendMessage(hWndList, LVM_INSERTCOLUMNW, LVI_SCRIPT_FILE, (LPARAM)&lvc);

    lvc.mask=LVCF_TEXT|LVCF_WIDTH|LVCF_SUBITEM;
    lvc.pszText=GetLangStringW(wLangSystem, STRID_HOTKEY);
    lvc.cx=nColumnWidth2;
    lvc.iSubItem=LVI_SCRIPT_HOTKEY;
    SendMessage(hWndList, LVM_INSERTCOLUMNW, LVI_SCRIPT_HOTKEY, (LPARAM)&lvc);

    FillScriptListW(hWndList);
  }
  else if (uMsg == WM_NOTIFY)
  {
    if (wParam == IDC_SCRIPTS_LIST)
    {
      if (((NMLISTVIEW *)lParam)->hdr.code == LVN_ITEMCHANGED)
      {
        if (((NMLISTVIEW *)lParam)->uNewState & LVIS_STATEIMAGEMASK)
        {
          BOOL bNewState;
          BOOL bOldState;

          bNewState=((((NMLISTVIEW *)lParam)->uNewState & LVIS_STATEIMAGEMASK) >> 12) - 1;
          bOldState=((((NMLISTVIEW *)lParam)->uOldState & LVIS_STATEIMAGEMASK) >> 12) - 1;

          if (bNewState >=0 && bOldState >=0 && bNewState != bOldState)
            bListChanged=TRUE;
        }
        if (((NMLISTVIEW *)lParam)->uNewState & LVIS_SELECTED)
        {
          SendMessage(hWndHotkey, HKM_SETHOTKEY, ((NMLISTVIEW *)lParam)->lParam, 0);
          EnableWindow(hWndExecButton, TRUE);
          EnableWindow(hWndEditButton, TRUE);
          EnableWindow(hWndHotkey, TRUE);
          EnableWindow(hWndAssignButton, TRUE);
          nSelItem=((NMLISTVIEW *)lParam)->iItem;
        }
        if (((NMLISTVIEW *)lParam)->uOldState & LVIS_SELECTED)
        {
          SendMessage(hWndHotkey, HKM_SETHOTKEY, 0, 0);
          EnableWindow(hWndExecButton, FALSE);
          EnableWindow(hWndEditButton, FALSE);
          EnableWindow(hWndHotkey, FALSE);
          EnableWindow(hWndAssignButton, FALSE);
          nSelItem=-1;
        }
      }
      else if ((int)((NMHDR *)lParam)->code == NM_DBLCLK) PostMessage(hDlg, WM_COMMAND, IDC_EXEC, 0);
    }
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDC_ASSIGN)
    {
      LVITEMW lvi;
      WORD wHotkey;
      wchar_t wszFile[MAX_PATH];
      wchar_t wszHotkey[MAX_PATH];

      //Get macro name
      lvi.mask=LVIF_TEXT;
      lvi.pszText=wszFile;
      lvi.cchTextMax=MAX_PATH;
      lvi.iItem=nSelItem;
      lvi.iSubItem=LVI_SCRIPT_FILE;
      SendMessage(hWndList, LVM_GETITEMW, 0, (LPARAM)&lvi);

      //Get macro hotkey
      wHotkey=(WORD)SendMessage(hWndHotkey, HKM_GETHOTKEY, 0, 0);

      //Register hotkey
      if (RegisterHotkeyW(wszFile, wHotkey))
      {
        GetHotkeyStringW(wHotkey, wszHotkey);

        lvi.mask=LVIF_PARAM;
        lvi.iItem=nSelItem;
        lvi.iSubItem=LVI_SCRIPT_FILE;
        lvi.lParam=wHotkey;
        SendMessage(hWndList, LVM_SETITEMW, 0, (LPARAM)&lvi);

        lvi.mask=LVIF_TEXT;
        lvi.pszText=wszHotkey;
        lvi.iItem=nSelItem;
        lvi.iSubItem=LVI_SCRIPT_HOTKEY;
        SendMessage(hWndList, LVM_SETITEMW, 0, (LPARAM)&lvi);

        SendMessage(hWndHotkey, HKM_SETHOTKEY, wHotkey, 0);
        bListChanged=TRUE;
      }
      else SetFocus(hWndHotkey);

      return TRUE;
    }
    else if (LOWORD(wParam) == IDC_SETTINGS)
    {
      DialogBoxW(hInstanceDLL, MAKEINTRESOURCEW(IDD_SETTINGS), hDlg, (DLGPROC)SettingsDlgProcW);
    }
    else if (LOWORD(wParam) == IDOK ||
             LOWORD(wParam) == IDC_EXEC ||
             LOWORD(wParam) == IDC_EDIT ||
             LOWORD(wParam) == IDCANCEL)
    {
      LVITEMW lvi;
      wchar_t wszName[MAX_PATH];
      int nWidth;

      nWidth=SendMessage(hWndList, LVM_GETCOLUMNWIDTH, LVI_SCRIPT_FILE, 0);
      if (nWidth != nColumnWidth1)
      {
        nColumnWidth1=nWidth;
        bMainDialogRectSave=TRUE;
      }
      nWidth=SendMessage(hWndList, LVM_GETCOLUMNWIDTH, LVI_SCRIPT_HOTKEY, 0);
      if (nWidth != nColumnWidth2)
      {
        nColumnWidth2=nWidth;
        bMainDialogRectSave=TRUE;
      }

      if (bListChanged || LOWORD(wParam) == IDOK)
      {
        SaveScriptListW(hWndList);
        SaveOptionsW(OF_MAIN|OF_WINDOWRECT);
      }

      if (LOWORD(wParam) == IDC_EXEC ||
          LOWORD(wParam) == IDC_EDIT)
      {
        //Get macro name
        lvi.mask=LVIF_TEXT;
        lvi.pszText=wszName;
        lvi.cchTextMax=MAX_PATH;
        lvi.iItem=nSelItem;
        lvi.iSubItem=LVI_SCRIPT_FILE;
        SendMessage(hWndList, LVM_GETITEMW, 0, (LPARAM)&lvi);
      }

      //End dialog
      EndDialog(hDlg, 0);

      //Execute
      if (LOWORD(wParam) == IDC_EXEC)
        ExecScriptW(wszName, NULL);
      else if (LOWORD(wParam) == IDC_EDIT)
        EditScriptW(wszName);
      return TRUE;
    }
  }
  else if (uMsg == WM_CLOSE)
  {
    PostMessage(hDlg, WM_COMMAND, IDCANCEL, 0);
    return TRUE;
  }

  if (DialogResizeMessages(&drs[0], &rcMainInitDialog, &rcMainCurrentDialog, hDlg, uMsg, wParam, lParam))
    bMainDialogRectSave=TRUE;

  return FALSE;
}

BOOL CALLBACK SettingsDlgProcA(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static HWND hWndModeManual;
  static HWND hWndModeAuto;
  static HWND hWndModePortable;

  if (uMsg == WM_INITDIALOG)
  {
    SendMessage(hDlg, WM_SETICON, (WPARAM)ICON_BIG, (LPARAM)hIconDLL);
    hWndModeManual=GetDlgItem(hDlg, IDC_SETTINGS_MODEMANUAL);
    hWndModeAuto=GetDlgItem(hDlg, IDC_SETTINGS_MODEAUTO);
    hWndModePortable=GetDlgItem(hDlg, IDC_SETTINGS_MODEPORTABLE);

    SetWindowTextA(hDlg, GetLangStringA(wLangSystem, STRID_SETTINGS));
    SetDlgItemTextA(hDlg, IDC_SETTINGS_MODE_GROUP, GetLangStringA(wLangSystem, STRID_MODE));
    SetDlgItemTextA(hDlg, IDC_SETTINGS_MODEMANUAL, GetLangStringA(wLangSystem, STRID_MODEMANUAL));
    SetDlgItemTextA(hDlg, IDC_SETTINGS_MODEAUTO, GetLangStringA(wLangSystem, STRID_MODEAUTO));
    SetDlgItemTextA(hDlg, IDC_SETTINGS_MODEPORTABLE, GetLangStringA(wLangSystem, STRID_MODEPORTABLE));
    SetDlgItemTextA(hDlg, IDOK, GetLangStringA(wLangSystem, STRID_OK));
    SetDlgItemTextA(hDlg, IDCANCEL, GetLangStringA(wLangSystem, STRID_CANCEL));

    if (nRegisterMode == RM_MANUAL)
      SendMessage(hWndModeManual, BM_SETCHECK, BST_CHECKED, 0);
    else if (nRegisterMode == RM_AUTO)
      SendMessage(hWndModeAuto, BM_SETCHECK, BST_CHECKED, 0);
    else if (nRegisterMode == RM_PORTABLE)
      SendMessage(hWndModePortable, BM_SETCHECK, BST_CHECKED, 0);
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDOK)
    {
      if (SendMessage(hWndModeManual, BM_GETCHECK, 0, 0) == BST_CHECKED)
        nRegisterMode=RM_MANUAL;
      else if (SendMessage(hWndModeAuto, BM_GETCHECK, 0, 0) == BST_CHECKED)
        nRegisterMode=RM_AUTO;
      else if (SendMessage(hWndModePortable, BM_GETCHECK, 0, 0) == BST_CHECKED)
        nRegisterMode=RM_PORTABLE;

      SaveOptionsA(OF_MAIN);

      EndDialog(hDlg, 0);
    }
    else if (LOWORD(wParam) == IDCANCEL)
    {
      EndDialog(hDlg, 0);
    }
  }
  else if (uMsg == WM_CLOSE)
  {
    PostMessage(hDlg, WM_COMMAND, IDCANCEL, 0);
    return TRUE;
  }
  return FALSE;
}

BOOL CALLBACK SettingsDlgProcW(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static HWND hWndModeManual;
  static HWND hWndModeAuto;
  static HWND hWndModePortable;

  if (uMsg == WM_INITDIALOG)
  {
    SendMessage(hDlg, WM_SETICON, (WPARAM)ICON_BIG, (LPARAM)hIconDLL);
    hWndModeManual=GetDlgItem(hDlg, IDC_SETTINGS_MODEMANUAL);
    hWndModeAuto=GetDlgItem(hDlg, IDC_SETTINGS_MODEAUTO);
    hWndModePortable=GetDlgItem(hDlg, IDC_SETTINGS_MODEPORTABLE);

    SetWindowTextW(hDlg, GetLangStringW(wLangSystem, STRID_SETTINGS));
    SetDlgItemTextW(hDlg, IDC_SETTINGS_MODE_GROUP, GetLangStringW(wLangSystem, STRID_MODE));
    SetDlgItemTextW(hDlg, IDC_SETTINGS_MODEMANUAL, GetLangStringW(wLangSystem, STRID_MODEMANUAL));
    SetDlgItemTextW(hDlg, IDC_SETTINGS_MODEAUTO, GetLangStringW(wLangSystem, STRID_MODEAUTO));
    SetDlgItemTextW(hDlg, IDC_SETTINGS_MODEPORTABLE, GetLangStringW(wLangSystem, STRID_MODEPORTABLE));
    SetDlgItemTextW(hDlg, IDOK, GetLangStringW(wLangSystem, STRID_OK));
    SetDlgItemTextW(hDlg, IDCANCEL, GetLangStringW(wLangSystem, STRID_CANCEL));

    if (nRegisterMode == RM_MANUAL)
      SendMessage(hWndModeManual, BM_SETCHECK, BST_CHECKED, 0);
    else if (nRegisterMode == RM_AUTO)
      SendMessage(hWndModeAuto, BM_SETCHECK, BST_CHECKED, 0);
    else if (nRegisterMode == RM_PORTABLE)
      SendMessage(hWndModePortable, BM_SETCHECK, BST_CHECKED, 0);
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDOK)
    {
      if (SendMessage(hWndModeManual, BM_GETCHECK, 0, 0) == BST_CHECKED)
        nRegisterMode=RM_MANUAL;
      else if (SendMessage(hWndModeAuto, BM_GETCHECK, 0, 0) == BST_CHECKED)
        nRegisterMode=RM_AUTO;
      else if (SendMessage(hWndModePortable, BM_GETCHECK, 0, 0) == BST_CHECKED)
        nRegisterMode=RM_PORTABLE;

      SaveOptionsW(OF_MAIN);

      EndDialog(hDlg, 0);
    }
    else if (LOWORD(wParam) == IDCANCEL)
    {
      EndDialog(hDlg, 0);
    }
  }
  else if (uMsg == WM_CLOSE)
  {
    PostMessage(hDlg, WM_COMMAND, IDCANCEL, 0);
    return TRUE;
  }
  return FALSE;
}

SCRIPTINFOA* StackInsertScriptA(HSTACK *hStack)
{
  SCRIPTINFOA *lpElement=NULL;

  StackInsertIndex((stack **)&hStack->first, (stack **)&hStack->last, (stack **)&lpElement, -1, sizeof(SCRIPTINFOA));
  return lpElement;
}

SCRIPTINFOW* StackInsertScriptW(HSTACK *hStack)
{
  SCRIPTINFOW *lpElement=NULL;

  StackInsertIndex((stack **)&hStack->first, (stack **)&hStack->last, (stack **)&lpElement, -1, sizeof(SCRIPTINFOW));
  return lpElement;
}

SCRIPTINFOA* StackGetScriptA(HSTACK *hStack, int nIndex)
{
  SCRIPTINFOA *lpElement=(SCRIPTINFOA *)hStack->first;
  int i;

  if (nIndex == -1) return NULL;

  for (i=0; lpElement; ++i)
  {
    if (i == nIndex)
      break;

    lpElement=lpElement->next;
  }
  return lpElement;
}

SCRIPTINFOW* StackGetScriptW(HSTACK *hStack, int nIndex)
{
  SCRIPTINFOW *lpElement=(SCRIPTINFOW *)hStack->first;
  int i;

  if (nIndex == -1) return NULL;

  for (i=0; lpElement; ++i)
  {
    if (i == nIndex)
      break;

    lpElement=lpElement->next;
  }
  return lpElement;
}

void StackFreeScript(HSTACK *hStack)
{
  StackClear((stack **)&hStack->first, (stack **)&hStack->last);
}

BOOL RegisterHotkeyA(char *pScriptName, WORD wHotkey)
{
  PLUGINFUNCTIONA *pfElement=NULL;
  char szPrefix[MAX_PATH];
  char szFunction[MAX_PATH];

  lstrcpyA(szPrefix, "Scripts::Main::");
  wsprintfA(szFunction, "%s%s", szPrefix, pScriptName);
  pfElement=(PLUGINFUNCTIONA *)SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)NULL, wHotkey);

  if (!pfElement)
  {
    if (pfElement=(PLUGINFUNCTIONA *)SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)szFunction, 0))
    {
      pfElement->wHotkey=wHotkey;

      if (!pfElement->wHotkey && !pfElement->bRunning && !pfElement->bOnStart)
        SendMessage(hMainWnd, AKD_DLLDELETE, 0, (LPARAM)pfElement);
    }
    else
    {
      if (wHotkey)
      {
        PLUGINFUNCTIONA pf={0};

        if (pfElement=(PLUGINFUNCTIONA *)SendMessage(hMainWnd, AKD_DLLADD, 0, (LPARAM)&pf))
        {
          lstrcpynA(pfElement->szFunction, szFunction, MAX_PATH);
          pfElement->nFunctionLen=lstrlenA(pfElement->szFunction);
          pfElement->wHotkey=wHotkey;
          pfElement->PluginProc=HotkeyProcA;
          pfElement->lpParameter=pfElement->szFunction + lstrlenA(szPrefix);
        }
      }
    }
    return TRUE;
  }
  else
  {
    wsprintfA(szBuffer, GetLangStringA(wLangSystem, STRID_HOTKEY_EXISTS), szFunction, pfElement->szFunction);
    MessageBoxA(hWndMainDlg, szBuffer, "Scripts plugin", MB_OK|MB_ICONEXCLAMATION);
    return FALSE;
  }
}

BOOL RegisterHotkeyW(wchar_t *wpScriptName, WORD wHotkey)
{
  PLUGINFUNCTIONW *pfElement=NULL;
  wchar_t wszPrefix[MAX_PATH];
  wchar_t wszFunction[MAX_PATH];

  lstrcpyW(wszPrefix, L"Scripts::Main::");
  wsprintfW(wszFunction, L"%s%s", wszPrefix, wpScriptName);
  pfElement=(PLUGINFUNCTIONW *)SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)NULL, wHotkey);

  if (!pfElement)
  {
    if (pfElement=(PLUGINFUNCTIONW *)SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)wszFunction, 0))
    {
      pfElement->wHotkey=wHotkey;

      if (!pfElement->wHotkey && !pfElement->bRunning && !pfElement->bOnStart)
        SendMessage(hMainWnd, AKD_DLLDELETE, 0, (LPARAM)pfElement);
    }
    else
    {
      if (wHotkey)
      {
        PLUGINFUNCTIONW pf={0};

        if (pfElement=(PLUGINFUNCTIONW *)SendMessage(hMainWnd, AKD_DLLADD, 0, (LPARAM)&pf))
        {
          lstrcpynW(pfElement->szFunction, wszFunction, MAX_PATH);
          pfElement->nFunctionLen=lstrlenW(pfElement->szFunction);
          pfElement->wHotkey=wHotkey;
          pfElement->PluginProc=HotkeyProcW;
          pfElement->lpParameter=pfElement->szFunction + lstrlenW(wszPrefix);
        }
      }
    }
    return TRUE;
  }
  else
  {
    wsprintfW(wszBuffer, GetLangStringW(wLangSystem, STRID_HOTKEY_EXISTS), wszFunction, pfElement->szFunction);
    MessageBoxW(hWndMainDlg, wszBuffer, L"Scripts plugin", MB_OK|MB_ICONEXCLAMATION);
    return FALSE;
  }
}

BOOL CALLBACK HotkeyProcA(void *lpParameter)
{
  char *pName=(char *)lpParameter;

  ExecScriptA(pName, NULL);
  return TRUE;
}

BOOL CALLBACK HotkeyProcW(void *lpParameter)
{
  wchar_t *wpName=(wchar_t *)lpParameter;

  ExecScriptW(wpName, NULL);
  return TRUE;
}

DWORD WINAPI ExecThreadProcA(LPVOID lpParameter)
{
  EXECSCRIPTA *es=(EXECSCRIPTA *)lpParameter;
  HANDLE hInitMutex=es->hInitMutex;
  HANDLE hExecMutex;
  HANDLE hThread=hExecThread;
  char *szScript=NULL;
  char *szArguments=NULL;
  int nScriptLen;
  int nArgumentsLen;

  if (es->pScript)
  {
    nScriptLen=lstrlenA(es->pScript);
    if (szScript=(char *)GlobalAlloc(GPTR, (nScriptLen + 1)))
      lstrcpynA(szScript, es->pScript, nScriptLen + 1);
  }
  if (es->pArguments)
  {
    nArgumentsLen=lstrlenA(es->pArguments);
    if (szArguments=(char *)GlobalAlloc(GPTR, (nArgumentsLen + 1)))
      lstrcpynA(szArguments, es->pArguments, nArgumentsLen + 1);
  }

  //Protect from double execution
  hExecMutex=CreateEventA(NULL, FALSE, FALSE, szMutexExecName);

  //Thread is initialized now unlock main thread
  SetEvent(hInitMutex);

  if (szScript && hExecMutex)
  {
    if (GetLastError() == ERROR_ALREADY_EXISTS)
      WaitForSingleObject(hExecMutex, INFINITE);

    //Run script
    {
      STARTUPINFOA si={0};
      PROCESS_INFORMATION pi={0};

      wsprintfA(szCmdLine, "WScript.exe \"%s\\%s\" ", szScriptsDir, szScript);
      if (szArguments) lstrcatA(szCmdLine, szArguments);
      si.cb=sizeof(STARTUPINFO);
      si.wShowWindow=SW_SHOWNORMAL;

      if (CreateProcessA(NULL, szCmdLine, NULL, NULL, FALSE, 0, NULL, NULL, &si, &pi))
      {
        WaitForSingleObject(pi.hProcess, INFINITE);
        CloseHandle(pi.hProcess);
        CloseHandle(pi.hThread);
      }
    }

    SetEvent(hExecMutex);
    CloseHandle(hExecMutex);
  }

  //Free thread handle
  if (hThread)
  {
    CloseHandle(hThread);
    hThread=NULL;
  }
  return 0;
}

DWORD WINAPI ExecThreadProcW(LPVOID lpParameter)
{
  EXECSCRIPTW *es=(EXECSCRIPTW *)lpParameter;
  HANDLE hInitMutex=es->hInitMutex;
  HANDLE hExecMutex;
  HANDLE hThread=hExecThread;
  wchar_t *wszScript=NULL;
  wchar_t *wszArguments=NULL;
  int nScriptLen;
  int nArgumentsLen;

  if (es->wpScript)
  {
    nScriptLen=lstrlenW(es->wpScript);
    if (wszScript=(wchar_t *)GlobalAlloc(GPTR, (nScriptLen + 1) * sizeof(wchar_t)))
      lstrcpynW(wszScript, es->wpScript, nScriptLen + 1);
  }
  if (es->wpArguments)
  {
    nArgumentsLen=lstrlenW(es->wpArguments);
    if (wszArguments=(wchar_t *)GlobalAlloc(GPTR, (nArgumentsLen + 1) * sizeof(wchar_t)))
      lstrcpynW(wszArguments, es->wpArguments, nArgumentsLen + 1);
  }

  //Protect from double execution
  hExecMutex=CreateEventW(NULL, FALSE, FALSE, wszMutexExecName);

  //Thread is initialized now unlock main thread
  SetEvent(hInitMutex);

  if (wszScript && hExecMutex)
  {
    if (GetLastError() == ERROR_ALREADY_EXISTS)
      WaitForSingleObject(hExecMutex, INFINITE);

    //Run script
    {
      STARTUPINFOW si={0};
      PROCESS_INFORMATION pi={0};

      wsprintfW(wszCmdLine, L"WScript.exe \"%s\\%s\" ", wszScriptsDir, wszScript);
      if (wszArguments) lstrcatW(wszCmdLine, wszArguments);
      si.cb=sizeof(STARTUPINFOW);
      si.wShowWindow=SW_SHOWNORMAL;

      if (CreateProcessW(NULL, wszCmdLine, NULL, NULL, FALSE, 0, NULL, NULL, &si, &pi))
      {
        WaitForSingleObject(pi.hProcess, INFINITE);
        CloseHandle(pi.hProcess);
        CloseHandle(pi.hThread);
      }
    }

    SetEvent(hExecMutex);
    CloseHandle(hExecMutex);
  }
  if (wszArguments) GlobalFree((HGLOBAL)wszArguments);
  if (wszScript) GlobalFree((HGLOBAL)wszScript);

  //Free thread handle
  if (hThread)
  {
    CloseHandle(hThread);
    hThread=NULL;
  }
  return 0;
}

void FillScriptListA(HWND hWnd)
{
  WIN32_FIND_DATAA wfd;
  HANDLE hFind;
  char szFindFiles[MAX_PATH];
  char szHotkey[MAX_PATH];
  char *pExt;
  LVITEMA lvi;
  int nIndex;
  int i=0;

  //Script files
  wsprintfA(szFindFiles, "%s\\*.*", szScriptsDir);

  if ((hFind=FindFirstFileA(szFindFiles, &wfd)) != INVALID_HANDLE_VALUE)
  {
    do
    {
      if (wfd.cFileName[0] == '.' && (wfd.cFileName[1] == '\0' || (wfd.cFileName[1] == '.' && wfd.cFileName[2] == '\0'))) continue;

      pExt=GetFileExtA(wfd.cFileName);
      if (!lstrcmpiA(pExt, "js") ||
          !lstrcmpiA(pExt, "vbs") ||
          !lstrcmpiA(pExt, "wsh") ||
          !lstrcmpiA(pExt, "wsf"))
      {
        //Find hotkey
        {
          PLUGINFUNCTIONA *pfElement=NULL;

          lvi.mask=LVIF_TEXT;
          lvi.pszText=wfd.cFileName;
          lvi.iItem=i++;
          lvi.iSubItem=LVI_SCRIPT_FILE;
          nIndex=SendMessage(hWnd, LVM_INSERTITEMA, 0, (LPARAM)&lvi);

          wsprintfA(szBuffer, "Scripts::Main::%s", wfd.cFileName);
          pfElement=(PLUGINFUNCTIONA *)SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)szBuffer, 0);

          if (pfElement)
          {
            lvi.mask=LVIF_PARAM;
            lvi.iItem=nIndex;
            lvi.iSubItem=LVI_SCRIPT_FILE;
            lvi.lParam=pfElement->wHotkey;
            SendMessage(hWnd, LVM_SETITEMA, 0, (LPARAM)&lvi);

            GetHotkeyStringA(pfElement->wHotkey, szHotkey);
            lvi.mask=LVIF_TEXT;
            lvi.pszText=szHotkey;
            lvi.iItem=nIndex;
            lvi.iSubItem=LVI_SCRIPT_HOTKEY;
            SendMessage(hWnd, LVM_SETITEMA, 0, (LPARAM)&lvi);
          }
        }
      }
    }
    while (FindNextFileA(hFind, &wfd));

    FindClose(hFind);
  }
  lvi.stateMask=LVIS_SELECTED|LVIS_FOCUSED;
  lvi.state=LVIS_SELECTED|LVIS_FOCUSED;
  SendMessage(hWnd, LVM_SETITEMSTATE, 0, (LPARAM)&lvi);
}

void FillScriptListW(HWND hWnd)
{
  WIN32_FIND_DATAW wfd;
  HANDLE hFind;
  wchar_t wszFindFiles[MAX_PATH];
  wchar_t wszHotkey[MAX_PATH];
  wchar_t *wpExt;
  LVITEMW lvi;
  int nIndex;
  int i=0;

  //Script files
  wsprintfW(wszFindFiles, L"%s\\*.*", wszScriptsDir);

  if ((hFind=FindFirstFileW(wszFindFiles, &wfd)) != INVALID_HANDLE_VALUE)
  {
    do
    {
      if (wfd.cFileName[0] == '.' && (wfd.cFileName[1] == '\0' || (wfd.cFileName[1] == '.' && wfd.cFileName[2] == '\0'))) continue;

      wpExt=GetFileExtW(wfd.cFileName);
      if (!lstrcmpiW(wpExt, L"js") ||
          !lstrcmpiW(wpExt, L"vbs") ||
          !lstrcmpiW(wpExt, L"wsh") ||
          !lstrcmpiW(wpExt, L"wsf"))
      {
        //Find hotkey
        {
          PLUGINFUNCTIONW *pfElement=NULL;

          lvi.mask=LVIF_TEXT;
          lvi.pszText=wfd.cFileName;
          lvi.iItem=i++;
          lvi.iSubItem=LVI_SCRIPT_FILE;
          nIndex=SendMessage(hWnd, LVM_INSERTITEMW, 0, (LPARAM)&lvi);

          wsprintfW(wszBuffer, L"Scripts::Main::%s", wfd.cFileName);
          pfElement=(PLUGINFUNCTIONW *)SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)wszBuffer, 0);

          if (pfElement)
          {
            lvi.mask=LVIF_PARAM;
            lvi.iItem=nIndex;
            lvi.iSubItem=LVI_SCRIPT_FILE;
            lvi.lParam=pfElement->wHotkey;
            SendMessage(hWnd, LVM_SETITEMW, 0, (LPARAM)&lvi);

            GetHotkeyStringW(pfElement->wHotkey, wszHotkey);
            lvi.mask=LVIF_TEXT;
            lvi.pszText=wszHotkey;
            lvi.iItem=nIndex;
            lvi.iSubItem=LVI_SCRIPT_HOTKEY;
            SendMessage(hWnd, LVM_SETITEMW, 0, (LPARAM)&lvi);
          }
        }
      }
    }
    while (FindNextFileW(hFind, &wfd));

    FindClose(hFind);
  }
  lvi.stateMask=LVIS_SELECTED|LVIS_FOCUSED;
  lvi.state=LVIS_SELECTED|LVIS_FOCUSED;
  SendMessage(hWnd, LVM_SETITEMSTATE, 0, (LPARAM)&lvi);
}

void SaveScriptListA(HWND hWnd)
{
  PLUGINOPTIONA po;
  HANDLE hOptions;
  char szFile[MAX_PATH];
  LVITEMA lvi;
  int i=0;

  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_SAVE|POB_CLEAR, (LPARAM)szPluginName))
  {
    while (1)
    {
      //Get script name
      lvi.mask=LVIF_TEXT;
      lvi.pszText=szFile;
      lvi.cchTextMax=MAX_PATH;
      lvi.iItem=i;
      lvi.iSubItem=LVI_SCRIPT_FILE;
      if (!SendMessage(hWnd, LVM_GETITEMA, 0, (LPARAM)&lvi)) break;

      //Get script hotkey
      lvi.mask=LVIF_PARAM;
      lvi.iItem=i;
      lvi.iSubItem=LVI_SCRIPT_FILE;
      if (!SendMessage(hWnd, LVM_GETITEMA, 0, (LPARAM)&lvi)) break;

      //Save hotkey
      if (lvi.lParam)
      {
        po.pOptionName=szFile;
        po.dwType=PO_DWORD;
        po.lpData=(LPBYTE)&lvi.lParam;
        po.dwData=sizeof(DWORD);
        SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);
      }

      ++i;
    }
    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }
}

void SaveScriptListW(HWND hWnd)
{
  PLUGINOPTIONW po;
  HANDLE hOptions;
  wchar_t wszFile[MAX_PATH];
  LVITEMW lvi;
  int i=0;

  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_SAVE|POB_CLEAR, (LPARAM)wszPluginName))
  {
    while (1)
    {
      //Get script name
      lvi.mask=LVIF_TEXT;
      lvi.pszText=wszFile;
      lvi.cchTextMax=MAX_PATH;
      lvi.iItem=i;
      lvi.iSubItem=LVI_SCRIPT_FILE;
      if (!SendMessage(hWnd, LVM_GETITEMW, 0, (LPARAM)&lvi)) break;

      //Get script hotkey
      lvi.mask=LVIF_PARAM;
      lvi.iItem=i;
      lvi.iSubItem=LVI_SCRIPT_FILE;
      if (!SendMessage(hWnd, LVM_GETITEMW, 0, (LPARAM)&lvi)) break;

      //Save hotkey
      if (lvi.lParam)
      {
        po.pOptionName=wszFile;
        po.dwType=PO_DWORD;
        po.lpData=(LPBYTE)&lvi.lParam;
        po.dwData=sizeof(DWORD);
        SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);
      }

      ++i;
    }
    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }
}

void ExecScriptA(char *pScript, char *szArguments)
{
  EXECSCRIPTA es;
  HANDLE hInitMutex;

  if (hInitMutex=CreateEventA(NULL, FALSE, FALSE, szMutexInitName))
  {
    es.pScript=pScript;
    es.pArguments=szArguments;
    es.hInitMutex=hInitMutex;

    if (hExecThread=CreateThread(NULL, 0, ExecThreadProcA, &es, 0, &dwExecThreadId))
      WaitForSingleObject(hInitMutex, INFINITE);

    CloseHandle(hInitMutex);
  }
}

void ExecScriptW(wchar_t *wpScript, wchar_t *wszArguments)
{
  EXECSCRIPTW es;
  HANDLE hInitMutex;

  if (hInitMutex=CreateEventW(NULL, FALSE, FALSE, wszMutexInitName))
  {
    es.wpScript=wpScript;
    es.wpArguments=wszArguments;
    es.hInitMutex=hInitMutex;

    if (hExecThread=CreateThread(NULL, 0, ExecThreadProcW, &es, 0, &dwExecThreadId))
      WaitForSingleObject(hInitMutex, INFINITE);

    CloseHandle(hInitMutex);
  }
}

int EditScriptA(char *pScript)
{
  OPENDOCUMENTA od;
  COPYDATASTRUCT cds;
  char szFullName[MAX_PATH];

  wsprintfA(szFullName, "%s\\%s", szScriptsDir, pScript);

  od.hWnd=NULL;
  lstrcpyA(od.szFile, szFullName);
  od.szWorkDir[0]='\0';
  od.dwFlags=OD_ADT_BINARY_ERROR|OD_ADT_REG_CODEPAGE;
  od.nCodePage=0;
  od.bBOM=0;

  cds.dwData=CD_OPENDOCUMENT;
  cds.cbData=sizeof(OPENDOCUMENTA);
  cds.lpData=&od;
  return SendMessage(hMainWnd, WM_COPYDATA, (WPARAM)hMainWnd, (LPARAM)&cds);
}

int EditScriptW(wchar_t *wpScript)
{
  OPENDOCUMENTW od;
  COPYDATASTRUCT cds;
  wchar_t wszFullName[MAX_PATH];

  wsprintfW(wszFullName, L"%s\\%s", wszScriptsDir, wpScript);

  od.hWnd=NULL;
  lstrcpyW(od.szFile, wszFullName);
  od.szWorkDir[0]='\0';
  od.dwFlags=OD_ADT_BINARY_ERROR|OD_ADT_REG_CODEPAGE;
  od.nCodePage=0;
  od.bBOM=0;

  cds.dwData=CD_OPENDOCUMENT;
  cds.cbData=sizeof(OPENDOCUMENTW);
  cds.lpData=&od;
  return SendMessage(hMainWnd, WM_COPYDATA, (WPARAM)hMainWnd, (LPARAM)&cds);
}

void GetHotkeyStringA(WORD wHotkey, char *szString)
{
  char szKeyText[100];
  LONG lScan;

  szString[0]='\0';

  if (wHotkey)
  {
    if (HIBYTE(wHotkey) & HOTKEYF_CONTROL)
      lstrcatA(szString, "Ctrl + ");
    if (HIBYTE(wHotkey) & HOTKEYF_SHIFT)
      lstrcatA(szString, "Shift + ");
    if (HIBYTE(wHotkey) & HOTKEYF_ALT)
      lstrcatA(szString, "Alt + ");
    lScan=MapVirtualKey(LOBYTE(wHotkey), 0) << 16;
    if (HIBYTE(wHotkey) & HOTKEYF_EXT) lScan|=0x01000000L;
    GetKeyNameTextA(lScan, szKeyText, 100);
    lstrcatA(szString, szKeyText);
  }
}

void GetHotkeyStringW(WORD wHotkey, wchar_t *wszString)
{
  wchar_t wszKeyText[100];
  LONG lScan;

  wszString[0]='\0';

  if (wHotkey)
  {
    if (HIBYTE(wHotkey) & HOTKEYF_CONTROL)
      lstrcatW(wszString, L"Ctrl + ");
    if (HIBYTE(wHotkey) & HOTKEYF_SHIFT)
      lstrcatW(wszString, L"Shift + ");
    if (HIBYTE(wHotkey) & HOTKEYF_ALT)
      lstrcatW(wszString, L"Alt + ");
    lScan=MapVirtualKeyW(LOBYTE(wHotkey), 0) << 16;
    if (HIBYTE(wHotkey) & HOTKEYF_EXT) lScan|=0x01000000L;
    GetKeyNameTextW(lScan, wszKeyText, 100);
    lstrcatW(wszString, wszKeyText);
  }
}

char* GetFileNameA(char *pFile)
{
  int i;

  for (i=lstrlenA(pFile) - 1; i >= 0; --i)
  {
    if (pFile[i] == '\\')
      return (pFile + i + 1);
  }
  return pFile;
}

wchar_t* GetFileNameW(wchar_t *wpFile)
{
  int i;

  for (i=lstrlenW(wpFile) - 1; i >= 0; --i)
  {
    if (wpFile[i] == '\\')
      return (wpFile + i + 1);
  }
  return wpFile;
}

char* GetFileExtA(char *pFile)
{
  int i;

  for (i=lstrlenA(pFile) - 1; i >= 0; --i)
  {
    if (pFile[i] == '.') return (pFile + i + 1);
    else if (pFile[i] == '\\') break;
  }
  return "";
}

wchar_t* GetFileExtW(wchar_t *wpFile)
{
  int i;

  for (i=lstrlenW(wpFile) - 1; i >= 0; --i)
  {
    if (wpFile[i] == '.') return (wpFile + i + 1);
    else if (wpFile[i] == '\\') break;
  }
  return L"";
}

BOOL FileExistsA(char *pFile)
{
/*
  WIN32_FIND_DATAA wfdA;
  HANDLE hFind;

  if ((hFind=FindFirstFileA(pFile, &wfdA)) == INVALID_HANDLE_VALUE)
    return FALSE;

  FindClose(hFind);
  return TRUE;
*/
  if (GetFileAttributesA(pFile) == INVALID_FILE_ATTRIBUTES)
    return FALSE;
  return TRUE;
}

BOOL FileExistsW(wchar_t *wpFile)
{
/*
  WIN32_FIND_DATAW wfdW;
  HANDLE hFind;

  if ((hFind=FindFirstFileW(wpFile, &wfdW)) == INVALID_HANDLE_VALUE)
    return FALSE;

  FindClose(hFind);
  return TRUE;
*/
  if (GetFileAttributesW(wpFile) == INVALID_FILE_ATTRIBUTES)
    return FALSE;
  return TRUE;
}

BOOL GetWindowPos(HWND hWnd, HWND hWndOwner, RECT *rc)
{
  if (GetWindowRect(hWnd, rc))
  {
    rc->right-=rc->left;
    rc->bottom-=rc->top;

    if (hWndOwner)
    {
      if (!ScreenToClient(hWndOwner, (POINT *)&rc->left))
        return FALSE;
    }
    return TRUE;
  }
  return FALSE;
}

BOOL DialogResizeMessages(DIALOGRESIZE *drs, RECT *rcInit, RECT *rcCurrent, HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  if (uMsg == WM_INITDIALOG)
  {
    RECT rcTemplate;

    GetWindowPos(hDlg, NULL, rcInit);
    rcTemplate=*rcCurrent;
    *rcCurrent=*rcInit;

    if (rcTemplate.right && rcTemplate.bottom)
    {
      rcTemplate.left=rcInit->left + (rcInit->right - rcTemplate.right) / 2;
      rcTemplate.top=rcInit->top + (rcInit->bottom - rcTemplate.bottom) / 2;
      SetWindowPos(hDlg, 0, rcTemplate.left, rcTemplate.top, rcTemplate.right, rcTemplate.bottom, SWP_NOZORDER);
    }
  }
  else if (uMsg == WM_GETMINMAXINFO)
  {
    MINMAXINFO *mmi=(MINMAXINFO *)lParam;

    mmi->ptMinTrackSize.x=rcInit->right;
    mmi->ptMinTrackSize.y=rcInit->bottom;
  }
  else if (uMsg == WM_SIZE)
  {
    if (lParam)
    {
      RECT rcTemplate;
      RECT rcControl;
      DWORD dwFlags;
      POINT pt;
      int i;

      GetWindowPos(hDlg, NULL, &rcTemplate);
      pt.x=rcTemplate.right - rcCurrent->right;
      pt.y=rcTemplate.bottom - rcCurrent->bottom;
      *rcCurrent=rcTemplate;

      for (i=0; drs[i].lpWnd; ++i)
      {
        if (*drs[i].lpWnd)
        {
          dwFlags=0;
          if (drs[i].dwType & DRS_SIZE)
            dwFlags|=SWP_NOMOVE;
          else if (drs[i].dwType & DRS_MOVE)
            dwFlags|=SWP_NOSIZE;
          GetWindowPos(*drs[i].lpWnd, hDlg, &rcControl);
          SetWindowPos(*drs[i].lpWnd, 0, rcControl.left + ((drs[i].dwType & DRS_X)?pt.x:0), rcControl.top + ((drs[i].dwType & DRS_Y)?pt.y:0), rcControl.right + ((drs[i].dwType & DRS_X)?pt.x:0), rcControl.bottom + ((drs[i].dwType & DRS_Y)?pt.y:0), dwFlags|SWP_NOZORDER);
        }
      }
      InvalidateRect(hDlg, NULL, TRUE);
      return TRUE;
    }
  }
  else if (uMsg == WM_PAINT)
  {
    PAINTSTRUCT ps;
    RECT rcGrip;
    HDC hDC;

    if (hDC=BeginPaint(hDlg, &ps))
    {
      GetClientRect(hDlg, &rcGrip);
      rcGrip.left=rcGrip.right - GetSystemMetrics(SM_CXVSCROLL);
      rcGrip.top=rcGrip.bottom - GetSystemMetrics(SM_CYVSCROLL);
      DrawFrameControl(hDC, &rcGrip, DFC_SCROLL, DFCS_SCROLLSIZEGRIP);
      EndPaint(hDlg, &ps);
    }
  }
  return FALSE;
}


//// Options

void ReadOptionsA()
{
  PLUGINOPTIONA po;
  HANDLE hOptions;

  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_READ, (LPARAM)szPluginName))
  {
    po.pOptionName="/RegisterMode";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&nRegisterMode;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="/WindowRect";
    po.dwType=PO_BINARY;
    po.lpData=(LPBYTE)&rcMainCurrentDialog;
    po.dwData=sizeof(RECT);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="/ColumnWidth1";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&nColumnWidth1;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="/ColumnWidth2";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&nColumnWidth2;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    //Scripts hotkeys
    {
      WIN32_FIND_DATAA wfd;
      HANDLE hFind;
      char szFindFiles[MAX_PATH];
      char *pExt;

      wsprintfA(szFindFiles, "%s\\*.*", szScriptsDir);

      if ((hFind=FindFirstFileA(szFindFiles, &wfd)) != INVALID_HANDLE_VALUE)
      {
        do
        {
          if (wfd.cFileName[0] == '.' && (wfd.cFileName[1] == '\0' || (wfd.cFileName[1] == '.' && wfd.cFileName[2] == '\0'))) continue;

          pExt=GetFileExtA(wfd.cFileName);
          if (!lstrcmpiA(pExt, "js") ||
              !lstrcmpiA(pExt, "vbs") ||
              !lstrcmpiA(pExt, "wsh") ||
              !lstrcmpiA(pExt, "wsf"))
          {
            PLUGINOPTIONA po;
            DWORD dwHotkey=0;

            if (hOptions)
            {
              po.pOptionName=wfd.cFileName;
              po.dwType=PO_DWORD;
              po.lpData=(LPBYTE)&dwHotkey;
              po.dwData=sizeof(DWORD);
              SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);
            }
            if (dwHotkey) RegisterHotkeyA(wfd.cFileName, LOWORD(dwHotkey));
          }
        }
        while (FindNextFileA(hFind, &wfd));

        FindClose(hFind);
      }
    }
    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }
}

void ReadOptionsW()
{
  PLUGINOPTIONW po;
  HANDLE hOptions;

  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_READ, (LPARAM)wszPluginName))
  {
    po.pOptionName=L"/RegisterMode";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&nRegisterMode;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"/WindowRect";
    po.dwType=PO_BINARY;
    po.lpData=(LPBYTE)&rcMainCurrentDialog;
    po.dwData=sizeof(RECT);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"/ColumnWidth1";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&nColumnWidth1;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"/ColumnWidth2";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&nColumnWidth2;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    //Scripts hotkeys
    {
      WIN32_FIND_DATAW wfd;
      HANDLE hFind;
      wchar_t wszFindFiles[MAX_PATH];
      wchar_t *wpExt;

      wsprintfW(wszFindFiles, L"%s\\*.*", wszScriptsDir);

      if ((hFind=FindFirstFileW(wszFindFiles, &wfd)) != INVALID_HANDLE_VALUE)
      {
        do
        {
          if (wfd.cFileName[0] == '.' && (wfd.cFileName[1] == '\0' || (wfd.cFileName[1] == '.' && wfd.cFileName[2] == '\0'))) continue;

          wpExt=GetFileExtW(wfd.cFileName);
          if (!lstrcmpiW(wpExt, L"js") ||
              !lstrcmpiW(wpExt, L"vbs") ||
              !lstrcmpiW(wpExt, L"wsh") ||
              !lstrcmpiW(wpExt, L"wsf"))
          {
            PLUGINOPTIONW po;
            DWORD dwHotkey=0;

            if (hOptions)
            {
              po.pOptionName=wfd.cFileName;
              po.dwType=PO_DWORD;
              po.lpData=(LPBYTE)&dwHotkey;
              po.dwData=sizeof(DWORD);
              SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);
            }
            if (dwHotkey) RegisterHotkeyW(wfd.cFileName, LOWORD(dwHotkey));
          }
        }
        while (FindNextFileW(hFind, &wfd));

        FindClose(hFind);
      }
    }
    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }
}

void SaveOptionsA(DWORD dwFlags)
{
  PLUGINOPTIONA po;
  HANDLE hOptions;

  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_SAVE, (LPARAM)szPluginName))
  {
    if (dwFlags & OF_MAIN)
    {
      po.pOptionName="/RegisterMode";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&nRegisterMode;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);
    }

    if (dwFlags & OF_WINDOWRECT)
    {
      po.pOptionName="/WindowRect";
      po.dwType=PO_BINARY;
      po.lpData=(LPBYTE)&rcMainCurrentDialog;
      po.dwData=sizeof(RECT);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName="/ColumnWidth1";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&nColumnWidth1;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName="/ColumnWidth2";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&nColumnWidth2;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);
    }

    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }
}

void SaveOptionsW(DWORD dwFlags)
{
  PLUGINOPTIONW po;
  HANDLE hOptions;

  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_SAVE, (LPARAM)wszPluginName))
  {
    if (dwFlags & OF_MAIN)
    {
      po.pOptionName=L"/RegisterMode";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&nRegisterMode;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);
    }

    if (dwFlags & OF_WINDOWRECT)
    {
      po.pOptionName=L"/WindowRect";
      po.dwType=PO_BINARY;
      po.lpData=(LPBYTE)&rcMainCurrentDialog;
      po.dwData=sizeof(RECT);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName=L"/ColumnWidth1";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&nColumnWidth1;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName=L"/ColumnWidth2";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&nColumnWidth2;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);
    }

    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }
}

char* GetLangStringA(LANGID wLangID, int nStringID)
{
  static char szStringBuf[MAX_PATH];

  WideCharToMultiByte(CP_ACP, 0, GetLangStringW(wLangID, nStringID), -1, szStringBuf, MAX_PATH, NULL, NULL);
  return szStringBuf;
}

wchar_t* GetLangStringW(LANGID wLangID, int nStringID)
{
  if (wLangID == LANG_RUSSIAN)
  {
    if (nStringID == STRID_ARCHITECTURE_MISMATCH)
      return L"   AkelPad'  .";
    if (nStringID == STRID_HOTKEY_EXISTS)
      return L"%s\n\n     \"%s\".";
    if (nStringID == STRID_WINDOW_NOT_FOUND)
      return L" AkelPad'  .";
    if (nStringID == STRID_SCRIPT)
      return L"";
    if (nStringID == STRID_HOTKEY)
      return L" ";
    if (nStringID == STRID_EXEC)
      return L"";
    if (nStringID == STRID_EDIT)
      return L"";
    if (nStringID == STRID_ASSIGN)
      return L"";
    if (nStringID == STRID_SETTINGS)
      return L"...";
    if (nStringID == STRID_MODE)
      return L"  Scripts.dll";
    if (nStringID == STRID_MODEMANUAL)
      return L" ";
    if (nStringID == STRID_MODEAUTO)
      return L" ";
    if (nStringID == STRID_MODEPORTABLE)
      return L" ";
    if (nStringID == STRID_OK)
      return L"OK";
    if (nStringID == STRID_CANCEL)
      return L"";
  }
  else
  {
    if (nStringID == STRID_ARCHITECTURE_MISMATCH)
      return L"Architecture of the plugin and the AkelPad aren't compatible.";
    if (nStringID == STRID_HOTKEY_EXISTS)
      return L"%s\n\nHotkey already assigned to \"%s\".";
    if (nStringID == STRID_WINDOW_NOT_FOUND)
      return L"AkelPad's window not found.";
    if (nStringID == STRID_SCRIPT)
      return L"Script";
    if (nStringID == STRID_HOTKEY)
      return L"Hotkey";
    if (nStringID == STRID_EXEC)
      return L"Exec";
    if (nStringID == STRID_EDIT)
      return L"Edit";
    if (nStringID == STRID_ASSIGN)
      return L"Assign";
    if (nStringID == STRID_SETTINGS)
      return L"Settings...";
    if (nStringID == STRID_MODE)
      return L"Register Scripts.dll mode";
    if (nStringID == STRID_MODEMANUAL)
      return L"Manual mode";
    if (nStringID == STRID_MODEAUTO)
      return L"Automatic mode";
    if (nStringID == STRID_MODEPORTABLE)
      return L"Portable mode";
    if (nStringID == STRID_OK)
      return L"OK";
    if (nStringID == STRID_CANCEL)
      return L"Cancel";
  }
  return L"";
}

BOOL IsExtCallParamValid(LPARAM lParam, int nIndex)
{
  if (*((int *)lParam) >= (int)((nIndex + 1) * sizeof(int)))
    return TRUE;
  return FALSE;
}

int GetExtCallParam(LPARAM lParam, int nIndex)
{
  if (*((int *)lParam) >= (int)((nIndex + 1) * sizeof(int)))
    return *(((int *)lParam) + nIndex);
  return 0;
}

void InitCommon(PLUGINDATA *pd)
{
  bInitCommon=TRUE;
  hMainWnd=pd->hMainWnd;
  bOldWindows=pd->bOldWindows;

  if (bOldWindows)
  {
    int i;

    for (i=0; ((char *)pd->pFunction)[i] != ':'; ++i)
      szPluginName[i]=((char *)pd->pFunction)[i];
    szPluginName[i]='\0';
  }
  else
  {
    int i;

    for (i=0; ((wchar_t *)pd->pFunction)[i] != ':'; ++i)
      wszPluginName[i]=((wchar_t *)pd->pFunction)[i];
    wszPluginName[i]='\0';
  }

  if (bOldWindows)
  {
    wsprintfA(szScriptsDir, "%s\\AkelFiles\\Plugs\\Scripts", (char *)pd->pAkelDir);
    ReadOptionsA();
  }
  else
  {
    wsprintfW(wszScriptsDir, L"%s\\AkelFiles\\Plugs\\Scripts", (wchar_t *)pd->pAkelDir);
    ReadOptionsW();
  }
}

void InitMain()
{
  bInitMain=TRUE;

  //Register mode
  if (nRegisterMode == RM_AUTO || nRegisterMode == RM_PORTABLE)
  {
    const char szNameTypeLib[]="Scripts.dll";
    const char szNameProgID[]="AkelPad.document";
    const char szObjectDescription[]="AkelPad COM component";
    char szDLL[MAX_PATH];

    if (IsServerRegisteredA(hInstanceDLL, (REFCLSID)&CLSID_IDocument, szNameTypeLib, szNameProgID, szObjectDescription, szServerFile, MAX_PATH) == S_OK)
    {
      GetModuleFileNameA(hInstanceDLL, szDLL, MAX_PATH);

      if (lstrcmpiA(szServerFile, szDLL))
      {
        DllRegisterServer();
        bServerRegistered=TRUE;
      }
      else bServerRegistered=FALSE;
    }
    else
    {
      DllRegisterServer();
      bServerRegistered=TRUE;
    }
  }
}

void UninitMain()
{
  bInitMain=FALSE;

  if (bOldWindows)
  {
    if (bMainDialogRectSave) SaveOptionsA(OF_WINDOWRECT);
  }
  else
  {
    if (bMainDialogRectSave) SaveOptionsW(OF_WINDOWRECT);
  }
  StackFreeScript(&hScriptsStack);

  //Register mode
  if (nRegisterMode == RM_PORTABLE)
  {
    if (bServerRegistered)
    {
      DllUnregisterServer();
      bServerRegistered=FALSE;

      if (szServerFile[0])
      {
        STARTUPINFOA si={0};
        PROCESS_INFORMATION pi={0};

        wsprintfA(szCmdLine, "regsvr32.exe /s \"%s\" ", szServerFile);
        si.cb=sizeof(STARTUPINFOA);
        si.wShowWindow=SW_SHOWNORMAL;

        if (CreateProcessA(NULL, szCmdLine, NULL, NULL, FALSE, 0, NULL, NULL, &si, &pi))
        {
          CloseHandle(pi.hProcess);
          CloseHandle(pi.hThread);
        }
      }
    }
  }
}


//Entry point
BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
  if (fdwReason == DLL_PROCESS_ATTACH)
  {
    hInstanceDLL=hinstDLL;
    g_DocumentTypeInfo=NULL;
    g_nObjs=g_nLocks=0;
    MyIClassFactoryObj.lpVtbl=(IClassFactoryVtbl *)&MyIClassFactoryVtbl;

    if (!GetWindowsDirectoryW(NULL, 0))
    {
      bOldWindows=TRUE;
      hIconDLL=LoadIconA(hinstDLL, MAKEINTRESOURCEA(IDI_ICON_SCRIPTS));
    }
    else
    {
      bOldWindows=FALSE;
      hIconDLL=LoadIconW(hinstDLL, MAKEINTRESOURCEW(IDI_ICON_SCRIPTS));
    }
    GetMainInfo();
    DisableThreadLibraryCalls(hinstDLL);
  }
  else if (fdwReason == DLL_THREAD_ATTACH)
  {
  }
  else if (fdwReason == DLL_THREAD_DETACH)
  {
  }
  else if (fdwReason == DLL_PROCESS_DETACH)
  {
    if (hIconDLL) DestroyIcon(hIconDLL);
    if (hAkelProcess) CloseHandle(hAkelProcess);
    if (g_DocumentTypeInfo) g_DocumentTypeInfo->lpVtbl->Release(g_DocumentTypeInfo);
    if (bInitMain) UninitMain();
  }
  return TRUE;
}
