#ifndef _SCRIPTS_H_
#define _SCRIPTS_H_

extern PLUGINDATA pdMainInfo;
extern HWND hMainWnd;
extern BOOL bOldWindows;
extern LANGID wLangSystem;
extern BOOL bInitialize;
extern HINSTANCE hInstanceDLL;
extern HINSTANCE hAkelInstance;
extern HANDLE hAkelProcess;
extern DWORD dwAkelProcessId;
extern DWORD dwArchitectVersion;
extern HWND hWndDialog;
extern HICON hIconDLL;
extern char szScriptsDir[MAX_PATH];
extern wchar_t wszScriptsDir[MAX_PATH];
extern char szBuffer[1024];
extern wchar_t wszBuffer[1024];
extern char szCmdLine[32768];
extern wchar_t wszCmdLine[32768];
extern HSTACK hScriptsStack;

//Defines
#define DLLA_SCRIPTS_EXEC  1
#define DLLA_SCRIPTS_EDIT  2

#define STRID_ARCHITECTURE_MISMATCH 1
#define STRID_HOTKEY_EXISTS         2
#define STRID_WINDOW_NOT_FOUND      3
#define STRID_SCRIPT                4
#define STRID_HOTKEY                5
#define STRID_EXEC                  6
#define STRID_EDIT                  7
#define STRID_ASSIGN                8
#define STRID_SETTINGS              9
#define STRID_MODE                  10
#define STRID_MODEMANUAL            11
#define STRID_MODEAUTO              12
#define STRID_MODEPORTABLE          13
#define STRID_OK                    14
#define STRID_CANCEL                15

#define OF_MAIN            0x1
#define OF_WINDOWRECT      0x2

#define RM_MANUAL   0
#define RM_AUTO     1
#define RM_PORTABLE 2

#define LVI_SCRIPT_FILE         0
#define LVI_SCRIPT_HOTKEY       1

#define BUFFER_SIZE             1024

#define APP_MAIN_CLASS_3A       "AkelPad3"
#define APP_MAIN_CLASS_3W      L"AkelPad3"
#define APP_MAIN_CLASS_4A       "AkelPad4"
#define APP_MAIN_CLASS_4W      L"AkelPad4"

//DIALOGRESIZE
#define DRS_SIZE  0x1 //Resize control
#define DRS_MOVE  0x2 //Move control
#define DRS_X     0x4 //X value
#define DRS_Y     0x8 //Y value

#ifndef INVALID_FILE_ATTRIBUTES
  #define INVALID_FILE_ATTRIBUTES ((DWORD)-1)
#endif
#ifndef IMF_AUTOFONTSIZEADJUST
  #define IMF_AUTOFONTSIZEADJUST 0x0010
#endif

typedef struct {
  char *pScript;
  char *pArguments;
  HANDLE hInitMutex;
} EXECSCRIPTA;

typedef struct {
  wchar_t *wpScript;
  wchar_t *wpArguments;
  HANDLE hInitMutex;
} EXECSCRIPTW;

typedef struct _SCRIPTINFOA {
  struct _SCRIPTINFOA *next;
  struct _SCRIPTINFOA *prev;
  char szScriptFile[MAX_PATH];
  WORD wHotkey;
  PLUGINFUNCTIONA *pf;
} SCRIPTINFOA;

typedef struct _SCRIPTINFOW {
  struct _SCRIPTINFOW *next;
  struct _SCRIPTINFOW *prev;
  wchar_t wszScriptFile[MAX_PATH];
  WORD wHotkey;
  PLUGINFUNCTIONW *pf;
} SCRIPTINFOW;

typedef struct {
  HWND *lpWnd;   //Control window
  DWORD dwType;  //See DRS_* defines
} DIALOGRESIZE;

#ifndef LVM_SETEXTENDEDLISTVIEWSTYLE
  #define LVM_SETEXTENDEDLISTVIEWSTYLE (LVM_FIRST + 54)
#endif
#ifndef LVS_EX_GRIDLINES
  #define LVS_EX_GRIDLINES 0x00000001
#endif
#ifndef LVS_EX_FULLROWSELECT
  #define LVS_EX_FULLROWSELECT 0x00000020
#endif


//Functions prototypes
BOOL CALLBACK MainDlgProcA(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK MainDlgProcW(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK SettingsDlgProcA(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK SettingsDlgProcW(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
BOOL RegisterHotkeyA(char *pScriptName, WORD wHotkey);
BOOL RegisterHotkeyW(wchar_t *wpScriptName, WORD wHotkey);
BOOL CALLBACK HotkeyProcA(void *lpParameter);
BOOL CALLBACK HotkeyProcW(void *lpParameter);
DWORD WINAPI ExecThreadProcA(LPVOID lpParameter);
DWORD WINAPI ExecThreadProcW(LPVOID lpParameter);
void FillScriptListA(HWND hWnd);
void FillScriptListW(HWND hWnd);
void SaveScriptListA(HWND hWnd);
void SaveScriptListW(HWND hWnd);
void ExecScriptA(char *pScript, char *szArguments);
void ExecScriptW(wchar_t *wpScript, wchar_t *wszArguments);
int EditScriptA(char *pScript);
int EditScriptW(wchar_t *wpScript);
void GetHotkeyStringA(WORD wHotkey, char *szString);
void GetHotkeyStringW(WORD wHotkey, wchar_t *wszString);
char* GetFileNameA(char *pFile);
wchar_t* GetFileNameW(wchar_t *wpFile);
char* GetFileExtA(char *pFile);
wchar_t* GetFileExtW(wchar_t *wpFile);
BOOL FileExistsA(char *pFile);
BOOL FileExistsW(wchar_t *wpFile);
BOOL GetWindowPos(HWND hWnd, HWND hWndOwner, RECT *rc);
BOOL DialogResizeMessages(DIALOGRESIZE *drs, RECT *rcInit, RECT *rcCurrent, HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
void ReadOptionsA();
void ReadOptionsW();
void SaveOptionsA(DWORD dwFlags);
void SaveOptionsW(DWORD dwFlags);
char* GetLangStringA(LANGID wLangID, int nStringID);
wchar_t* GetLangStringW(LANGID wLangID, int nStringID);
BOOL IsExtCallParamValid(LPARAM lParam, int nIndex);
int GetExtCallParam(LPARAM lParam, int nIndex);
void InitCommon(PLUGINDATA *pd);
void InitMain();
void UninitMain();

#endif
